import os

from dataclasses import dataclass
from typing import List

import click

from linkml_runtime.utils.schemaview import SchemaView
from linkml.utils.generator import Generator, shared_arguments
from linkml._version import __version__
from linkml.utils.helpers import convert_to_snake_case
from openpyxl import Workbook
from openpyxl.worksheet.worksheet import Worksheet
from openpyxl.worksheet.datavalidation import DataValidation
from openpyxl.utils import get_column_letter


@dataclass
class ExcelGenerator(Generator):
    # ClassVars
    generatorname = os.path.basename(__file__)
    generatorversion = "0.1.1"
    valid_formats = ["xlsx"]
    uses_schemaloader = False
    requires_metamodel = False

    def __post_init__(self) -> None:
        super().__post_init__()
        self.schemaview = SchemaView(self.schema)

    def create_workbook(self, workbook_name: str) -> Workbook:
        """
        Creates an Excel workbook using the openpyxl library and returns it.

        :param workbook_name: Name of the workbook to be created.
        :return: An openpyxl Workbook object representing the newly created workbook.
        """
        workbook = Workbook()
        workbook.title = workbook_name
        return workbook

    def get_workbook_name(self, workbook: Workbook) -> str:
        """
        Returns the name of the given workbook.

        :param workbook: The workbook whose name should be returned.
        :return: Name of the workbook.
        """
        return workbook.title

    def remove_worksheet_by_name(self, workbook: Workbook, worksheet_name: str) -> None:
        """
        Remove worksheet from workbook by name.
        """
        worksheet = workbook[worksheet_name]
        workbook.remove(worksheet)

    def create_worksheet(self, workbook: Workbook, worksheet_name: str) -> Worksheet:
        """
        Creates an Excel worksheet with the given name in the given workbook.

        :param workbook: The workbook to which the worksheet should be added.
        :param worksheet_name: Name of the worksheet to be created.
        """
        worksheet = workbook.create_sheet(worksheet_name)
        workbook_name = self.get_workbook_name(workbook)
        workbook.save(workbook_name)

        return worksheet

    def create_schema_worksheets(self, workbook: str) -> None:
        """
        Creates worksheets in a given Excel workbook based on the classes in the
        schema.

        :param workbook: The workbook to which the worksheet should be added.
        """
        sv = self.schemaview
        for cls_name, cls in sv.all_classes(imports=self.mergeimports).items():
            if not cls.mixin and not cls.abstract:
                self.create_worksheet(workbook, cls_name)

    def add_columns_to_worksheet(
        self, workbook: Workbook, worksheet_name: str, sheet_headings: List[str]
    ) -> None:
        """
        Get a worksheet by name and add a column to it in an existing workbook.

        :param workbook: The workbook to which the worksheet should be added.
        :param worksheet_name: Name of the worksheet to add the column to.
        :param column_data: List of data to populate the column with.
        """
        # Get the worksheet by name
        worksheet = workbook[worksheet_name]

        # Add the headings to the worksheet
        for i, heading in enumerate(sheet_headings):
            worksheet.cell(row=1, column=i + 1, value=heading)

        # Save the changes to the workbook
        workbook_name = self.get_workbook_name(workbook)
        workbook.save(workbook_name)

    def column_enum_validation(
        self,
        workbook: Workbook,
        worksheet_name: str,
        column_name: str,
        dropdown_values: List[str],
    ) -> None:
        """
        Get worksheet by name and add a dropdown to a specific column in it
        based on a list of values.

        :param workbook: The workbook to which the worksheet should be added.
        :param worksheet_name: Name of the worksheet to add the column dropdown to.
        :param column_name: Name of the worksheet column to add the dropdown to.
        :param dropdown_values: List of dropdown values to add to a column in a worksheet.
        """
        worksheet = workbook[worksheet_name]

        column_list = [cell.value for cell in worksheet[1]]
        column_number = column_list.index(column_name) + 1
        column_letter = get_column_letter(column_number)

        # Create the data validation object and set the dropdown values
        dv = DataValidation(
            type="list", formula1=f'"{",".join(dropdown_values)}"', allow_blank=True
        )

        worksheet.add_data_validation(dv)

        dv.add(f"{column_letter}2:{column_letter}1048576")

        workbook_name = self.get_workbook_name(workbook)
        workbook.save(workbook_name)

    def serialize(self, **kwargs) -> str:
        output = (
            convert_to_snake_case(self.schema.name) + ".xlsx"
            if not self.output
            else self.output
        )

        workbook = self.create_workbook(output)
        self.remove_worksheet_by_name(workbook, "Sheet")
        self.create_schema_worksheets(workbook)

        sv = self.schemaview
        for cls_name, cls in sv.all_classes(imports=self.mergeimports).items():
            if not cls.mixin and not cls.abstract:
                slots = [
                    s.name
                    for s in sv.class_induced_slots(cls_name, imports=self.mergeimports)
                ]
                self.add_columns_to_worksheet(workbook, cls_name, slots)

        enum_list = [
            e_name for e_name, _ in sv.all_enums(imports=self.mergeimports).items()
        ]
        for cls_name, cls in sv.all_classes(imports=self.mergeimports).items():
            if not cls.mixin and not cls.abstract:
                for s in sv.class_induced_slots(cls_name, imports=self.mergeimports):
                    if s.range in enum_list:
                        pv_list = []
                        for pv_name, _ in sv.get_enum(
                            s.range
                        ).permissible_values.items():
                            pv_list.append(pv_name)
                        self.column_enum_validation(workbook, cls_name, s.name, pv_list)


@shared_arguments(ExcelGenerator)
@click.command()
@click.option(
    "-o",
    "--output",
    type=click.Path(),
    help="""Name of Excel spreadsheet to be created""",
)
@click.version_option(__version__, "-V", "--version")
def cli(yamlfile, **kwargs):
    """Generate Excel representation of a LinkML model"""
    print(ExcelGenerator(yamlfile, **kwargs).serialize(**kwargs))


if __name__ == "__main__":
    cli()



import os
from collections import defaultdict
from copy import deepcopy
from dataclasses import field, dataclass
from typing import Dict, List, TextIO, Union, Optional, Set
from linkml.utils.ifabsent_functions import ifabsent_value_declaration

import click
from jinja2 import Template
# from linkml.generators import pydantic_GEN_VERSION
from linkml_runtime.linkml_model.meta import (Annotation, ClassDefinition,
                                              EnumDefinition,
                                              EnumDefinitionName,
                                              SchemaDefinition, TypeDefinition, AnonymousSlotExpression, SlotDefinition)
from linkml_runtime.utils.formatutils import camelcase, underscore
from linkml_runtime.utils.schemaview import SchemaView

from linkml.generators.common.type_designators import get_type_designator_value, get_accepted_type_designator_values
from linkml._version import __version__
from linkml.generators.oocodegen import OOCodeGenerator
from linkml.utils.generator import shared_arguments

default_template = """
{#-

  Jinja2 Template for a pydantic classes
-#}
from __future__ import annotations
from datetime import datetime, date
from enum import Enum
from typing import List, Dict, Optional, Any, Union, Literal
from pydantic import BaseModel as BaseModel, Field
from linkml_runtime.linkml_model import Decimal

metamodel_version = "{{metamodel_version}}"
version = "{{version if version else None}}"

class WeakRefShimBaseModel(BaseModel):
   __slots__ = '__weakref__'
    
class ConfiguredBaseModel(WeakRefShimBaseModel,
                validate_assignment = True, 
                validate_all = True, 
                underscore_attrs_are_private = True, 
                extra = {% if allow_extra %}'allow'{% else %}'forbid'{% endif %}, 
                arbitrary_types_allowed = True):
    pass                    

{% for e in enums.values() %}
class {{ e.name }}(str, Enum):
    {% if e.description -%}
    \"\"\"
    {{ e.description }}
    \"\"\"
    {%- endif %}
    {% for label, value in e['values'].items() -%}
    {{label}} = "{{value}}"
    {% endfor %}
    {% if not e['values'] -%}
    dummy = "dummy"
    {% endif %}
{% endfor %}

{%- for c in schema.classes.values() %}
class {{ c.name }} 
    {%- if class_isa_plus_mixins[c.name] -%}
        ({{class_isa_plus_mixins[c.name]|join(', ')}})
    {%- else -%}
        (ConfiguredBaseModel)
    {%- endif -%}                   
                  :
    {% if c.description -%}    
    \"\"\"
    {{ c.description }}
    \"\"\"
    {%- endif %}
    {% for attr in c.attributes.values() if c.attributes -%}
    {{attr.name}}: {{ attr.annotations['python_range'].value }} = Field(
    {%- if predefined_slot_values[c.name][attr.name] -%} 
        {{ predefined_slot_values[c.name][attr.name] }}
    {%- else -%}
        None
    {%- endif -%}    
    {%- if attr.title != None %}, title="{{attr.title}}"{% endif -%}
    {%- if attr.description %}, description=\"\"\"{{attr.description}}\"\"\"{% endif -%}
    {%- if attr.minimum_value != None %}, ge={{attr.minimum_value}}{% endif -%}
    {%- if attr.maximum_value != None %}, le={{attr.maximum_value}}{% endif -%}
    )
    {% else -%}
    None
    {% endfor %}

{% endfor %}

# Update forward refs
# see https://pydantic-docs.helpmanual.io/usage/postponed_annotations/
{% for c in schema.classes.values() -%} 
{{ c.name }}.update_forward_refs()
{% endfor %}
"""


def _get_pyrange(t: TypeDefinition, sv: SchemaView) -> str:
    pyrange = t.repr if t is not None else None
    if pyrange is None:
        pyrange = t.base
    if t.base == "XSDDateTime":
        pyrange = "datetime "
    if t.base == "XSDDate":
        pyrange = "date"
    if pyrange is None and t.typeof is not None:
        pyrange = _get_pyrange(sv.get_type(t.typeof), sv)
    if pyrange is None:
        raise Exception(f"No python type for range: {s.range} // {t}")
    return pyrange


@dataclass
class PydanticGenerator(OOCodeGenerator):
    """
    Generates Pydantic-compliant classes from a schema

    This is an alternative to the dataclasses-based Pythongen
    """

    # ClassVar overrides
    generatorname = os.path.basename(__file__)
    generatorversion = "0.0.1"
    valid_formats = ["pydantic"]

    # ObjectVars
    template_file: str = None
    allow_extra: bool = field(default_factory=lambda: False)
    gen_mixin_inheritance: bool = field(default_factory=lambda: True)

    # ObjectVars (identical to pythongen)
    gen_classvars: bool = field(default_factory=lambda: True)
    gen_slots: bool = field(default_factory=lambda: True)
    genmeta: bool = field(default_factory=lambda: False)
    emit_metadata: bool = field(default_factory=lambda: True)


    def generate_enums(
        self, all_enums: Dict[EnumDefinitionName, EnumDefinition]
    ) -> Dict[str, dict]:
        # TODO: make an explicit class to represent how an enum is passed to the template
        enums = {}
        for enum_name, enum_orignal in all_enums.items():
            enum = {"name": camelcase(enum_name), "values": {}}

            for pv in enum_orignal.permissible_values.values():
                label = self.generate_enum_label(pv.text)
                enum["values"][label] = pv.text

            enums[enum_name] = enum

        return enums

    def sort_classes(self, clist: List[ClassDefinition]) -> List[ClassDefinition]:
        """
        sort classes such that if C is a child of P then C appears after P in the list

        Overridden method include mixin classes

        TODO: This should move to SchemaView
        """
        clist = list(clist)
        slist = []  # sorted
        while len(clist) > 0:
            can_add = False
            for i in range(len(clist)):
                candidate = clist[i]
                can_add = False
                if candidate.is_a:
                    candidates = [candidate.is_a] + candidate.mixins
                else:
                    candidates = candidate.mixins
                if not candidates:
                    can_add = True
                else:
                    if set(candidates) <= set([p.name for p in slist]):
                        can_add = True
                if can_add:
                    slist = slist + [candidate]
                    del clist[i]
                    break
            if not can_add:
                raise ValueError(
                    f"could not find suitable element in {clist} that does not ref {slist}"
                )
        return slist

    def get_predefined_slot_values(self) -> Dict[str, Dict[str, str]]:
        """
        :return: Dictionary of dictionaries with predefined slot values for each class
        """
        sv = self.schemaview
        default_prefix = sv.schema.default_prefix
        slot_values = defaultdict(dict)
        for class_def in sv.all_classes().values():

            for slot_name in sv.class_slots(class_def.name):
                slot = sv.induced_slot(slot_name, class_def.name)
                if slot.designates_type:
                    target_value = get_type_designator_value(sv, slot, class_def)
                    slot_values[camelcase(class_def.name)][
                        slot.name
                    ] = f'"{target_value}"'
                    if slot.multivalued:
                        slot_values[camelcase(class_def.name)][slot.name] = (
                            "["
                            + slot_values[camelcase(class_def.name)][slot.name]
                            + "]"
                        )
                    slot_values[camelcase(class_def.name)][slot.name] = slot_values[camelcase(class_def.name)][slot.name]
                elif slot.ifabsent is not None:
                    value = ifabsent_value_declaration(slot.ifabsent, sv, class_def, slot)
                    slot_values[camelcase(class_def.name)][slot.name] = value
                # Multivalued slots that are either not inlined (just an identifier) or are
                # inlined as lists should get default_factory list, if they're inlined but
                # not as a list, that means a dictionary
                elif slot.multivalued:
                    has_identifier_slot = self.range_class_has_identifier_slot(slot)

                    if slot.inlined and not slot.inlined_as_list and has_identifier_slot:
                        slot_values[camelcase(class_def.name)][slot.name] = "default_factory=dict"
                    else:
                        slot_values[camelcase(class_def.name)][slot.name] = "default_factory=list"



        return slot_values

    def range_class_has_identifier_slot(self, slot):
        """
        Check if the range class of a slot has an identifier slot, via both slot.any_of and slot.range
        Should return False if the range is not a class, and also if the range is a class but has no
        identifier slot

        :param slot: SlotDefinition
        :return: bool
        """
        sv = self.schemaview
        has_identifier_slot = False
        if slot.any_of:
            for slot_range in slot.any_of:
                any_of_range = slot_range.range
                if any_of_range in sv.all_classes() and sv.get_identifier_slot(any_of_range) is not None:
                    has_identifier_slot = True
        if slot.range in sv.all_classes() and sv.get_identifier_slot(slot.range) is not None:
            has_identifier_slot = True
        return has_identifier_slot

    def get_class_isa_plus_mixins(self) -> Dict[str, List[str]]:
        """
        Generate the inheritance list for each class from is_a plus mixins
        :return:
        """
        sv = self.schemaview
        parents = {}
        for class_def in sv.all_classes().values():
            class_parents = []
            if class_def.is_a:
                class_parents.append(camelcase(class_def.is_a))
            if self.gen_mixin_inheritance and class_def.mixins:
                class_parents.extend([camelcase(mixin) for mixin in class_def.mixins])
            if len(class_parents) > 0:
                # Use the sorted list of classes to order the parent classes, but reversed to match MRO needs
                class_parents.sort(key=lambda x: self.sorted_class_names.index(x))
                class_parents.reverse()
                parents[camelcase(class_def.name)] = class_parents
        return parents

    def get_mixin_identifier_range(self, mixin) -> str:
        sv = self.schemaview
        id_ranges = list(
            {
                _get_pyrange(sv.get_type(sv.get_identifier_slot(c).range), sv)
                for c in sv.class_descendants(mixin.name, mixins=True)
                if sv.get_identifier_slot(c) is not None
            }
        )
        if len(id_ranges) == 0:
            return None
        elif len(id_ranges) == 1:
            return id_ranges[0]
        else:
            return f"Union[{'.'.join(id_ranges)}]"

    def get_class_slot_range(self, slot_range: str, inlined: bool, inlined_as_list: bool) -> str:
        sv = self.schemaview
        range_cls = sv.get_class(slot_range)

        # Hardcoded handling for Any
        if range_cls.class_uri == "linkml:Any":
            return "Any"

        # Inline the class itself only if the class is defined as inline, or if the class has no
        # identifier slot and also isn't a mixin.
        if inlined or inlined_as_list or (
            sv.get_identifier_slot(range_cls.name) is None
            and not sv.is_mixin(range_cls.name)
        ):
            if len([x for x in sv.class_induced_slots(slot_range) if x.designates_type]) > 0 and len(sv.class_descendants(slot_range)) > 1:
                return f"Union[" + ",".join([camelcase(c) for c in sv.class_descendants(slot_range)]) + "]"
            else:
                return f"{camelcase(slot_range)}"

        # For the more difficult cases, set string as the default and attempt to improve it
        range_cls_identifier_slot_range = "str"

        # For mixins, try to use the identifier slot of descendant classes
        if (
            self.gen_mixin_inheritance
            and sv.is_mixin(range_cls.name)
            and sv.get_identifier_slot(range_cls.name)
        ):
            range_cls_identifier_slot_range = self.get_mixin_identifier_range(range_cls)

        # If the class itself has an identifier slot, it can be allowed to overwrite a value from mixin above
        if (
            sv.get_identifier_slot(range_cls.name) is not None
            and sv.get_identifier_slot(range_cls.name).range is not None
        ):
            range_cls_identifier_slot_range = _get_pyrange(
                sv.get_type(sv.get_identifier_slot(range_cls.name).range), sv
            )

        return range_cls_identifier_slot_range

    def generate_python_range(self, slot_range, slot_def:SlotDefinition, class_def:ClassDefinition) -> str:
        """
        Generate the python range for a slot range value
        """
        sv = self.schemaview

        if slot_def.designates_type:
            pyrange = "Literal[" + ",".join(
                ["\"" + x + "\"" for x in get_accepted_type_designator_values(sv, slot_def, class_def)]) + "]"
        elif slot_range in sv.all_classes():
            pyrange = self.get_class_slot_range(slot_range, inlined=slot_def.inlined, inlined_as_list=slot_def.inlined_as_list)
        elif slot_range in sv.all_enums():
            pyrange = f"{camelcase(slot_range)}"
        elif slot_range in sv.all_types():
            t = sv.get_type(slot_range)
            pyrange = _get_pyrange(t, sv)
        elif slot_range is None:
            pyrange = "str"
        else:
            # TODO: default ranges in schemagen
            # pyrange = 'str'
            # logging.error(f'range: {s.range} is unknown')
            raise Exception(f"range: {slot_range}")
        return pyrange

    def generate_collection_key(self, slot_ranges: List[str], slot_def: SlotDefinition, class_def: ClassDefinition) -> Optional[str]:
        """
        Find the python range value (str, int, etc) for the identifier slot
        of a class used as a slot range.

        If a pyrange value matches a class name, the range of the identifier slot
        will be returned. If more than one match is found and they don't match,
        an exception will be raised.

        :param slot_ranges: list of python range values
        """

        collection_keys:Set[str] = set()

        if slot_ranges is None:
            return None

        for slot_range in slot_ranges:
            if slot_range is None or slot_range not in self.schemaview.all_classes():
                continue # ignore non-class ranges

            identifier_slot = self.schemaview.get_identifier_slot(slot_range)
            if identifier_slot is not None:
                collection_keys.add(self.generate_python_range(identifier_slot.range, slot_def, class_def))
        if len(collection_keys) > 1:
            raise Exception(f"Slot with any_of range has multiple identifier slot range types: {collection_keys}")
        if len(collection_keys) == 1:
            return list(collection_keys)[0]
        return None

    def serialize(self) -> str:
        sv = self.schemaview

        if self.template_file is not None:
            with open(self.template_file) as template_file:
                template_obj = Template(template_file.read())
        else:
            template_obj = Template(default_template)

        sv: SchemaView
        sv = self.schemaview
        schema = sv.schema
        pyschema = SchemaDefinition(
            id=schema.id,
            name=schema.name,
            description=schema.description.replace('"', '\\"') if schema.description else None,
        )
        enums = self.generate_enums(sv.all_enums())

        sorted_classes = self.sort_classes(list(sv.all_classes().values()))
        self.sorted_class_names = [camelcase(c.name) for c in sorted_classes]

        # Don't want to generate classes when class_uri is linkml:Any, will
        # just swap in typing.Any instead down below
        sorted_classes = [c for c in sorted_classes if c.class_uri != "linkml:Any"]

        for class_original in sorted_classes:
            class_def: ClassDefinition
            class_def = deepcopy(class_original)
            class_name = class_original.name
            class_def.name = camelcase(class_original.name)
            if class_def.is_a:
                class_def.is_a = camelcase(class_def.is_a)
            class_def.mixins = [camelcase(p) for p in class_def.mixins]
            if class_def.description:
                class_def.description = class_def.description.replace('"', '\\"')
            pyschema.classes[class_def.name] = class_def
            for attribute in list(class_def.attributes.keys()):
                del class_def.attributes[attribute]
            for sn in sv.class_slots(class_name):
                # TODO: fix runtime, copy should not be necessary
                s = deepcopy(sv.induced_slot(sn, class_name))
                # logging.error(f'Induced slot {class_name}.{sn} == {s.name} {s.range}')
                s.name = underscore(s.name)
                if s.description:
                    s.description = s.description.replace('"', '\\"')
                class_def.attributes[s.name] = s

                slot_ranges:List[str] = []

                if len(s.any_of) > 0 and s.range is not None:
                    raise ValueError("Slot cannot have both range and any_of defined")

                if s.any_of is not None and len(s.any_of) > 0:
                    # list comprehension here is pulling ranges from within AnonymousSlotExpression
                    slot_ranges.extend([r.range for r in s.any_of])
                else:
                    slot_ranges.append(s.range)

                pyranges = [self.generate_python_range(slot_range, s, class_def)
                            for slot_range in slot_ranges]

                pyranges = list(set(pyranges)) # remove duplicates
                pyranges.sort()

                if len(pyranges) == 1:
                    pyrange = pyranges[0]
                elif len(pyranges) > 1:
                    pyrange = f"Union[{', '.join(pyranges)}]"
                else:
                    raise Exception(f"Could not generate python range for {class_name}.{s.name}")

                if s.multivalued:
                    if s.inlined or s.inlined_as_list:
                        collection_key = self.generate_collection_key(slot_ranges, s, class_def)
                    else:
                        collection_key = None
                    if s.inlined == False or collection_key is None or s.inlined_as_list == True:
                        pyrange = f"List[{pyrange}]"
                    else:
                        pyrange = f"Dict[{collection_key}, {pyrange}]"
                if not s.required and not s.designates_type:
                    pyrange = f"Optional[{pyrange}]"
                ann = Annotation("python_range", pyrange)
                s.annotations[ann.tag] = ann
        code = template_obj.render(
            schema=pyschema,
            underscore=underscore,
            enums=enums,
            predefined_slot_values=self.get_predefined_slot_values(),
            allow_extra=self.allow_extra,
            metamodel_version=self.schema.metamodel_version,
            version=self.schema.version,
            class_isa_plus_mixins=self.get_class_isa_plus_mixins(),
        )
        return code

    def default_value_for_type(self, typ: str) -> str:
        return 'None'


@shared_arguments(PydanticGenerator)
@click.option(
    "--template_file", help="Optional jinja2 template to use for class generation"
)
@click.version_option(__version__, "-V", "--version")
@click.command()
def cli(
    yamlfile,
    template_file=None,
    head=True,
    emit_metadata=False,
    genmeta=False,
    classvars=True,
    slots=True,
    **args,
):
    """Generate pydantic classes to represent a LinkML model"""
    gen = PydanticGenerator(
        yamlfile,
        template_file=template_file,
        emit_metadata=head,
        genmeta=genmeta,
        gen_classvars=classvars,
        gen_slots=slots,
        **args,
    )
    print(gen.serialize())


if __name__ == "__main__":
    cli()



import abc
import re
import unicodedata
from dataclasses import dataclass, field
from typing import List, Optional, Dict

from linkml_runtime.linkml_model.meta import (ClassDefinition,
                                              SchemaDefinition, SlotDefinition,
                                              TypeDefinition)
from linkml_runtime.utils.formatutils import camelcase, lcamelcase, underscore
from linkml_runtime.utils.schemaview import SchemaView

from linkml.utils.generator import Generator

SAFE_NAME = str
TYPE_EXPRESSION = str
ANNOTATION = str
PACKAGE = str


@dataclass
class OODocument:
    """
    A collection of one or more OO classes
    """

    name: SAFE_NAME
    package: PACKAGE = None
    source_schema: SchemaDefinition = None
    classes: List["OOClass"] = field(default_factory=lambda: [])
    imports: List[str] = field(default_factory=lambda: [])


@dataclass
class OOField:
    """
    A field belonging to an OO class that corresponds to a LinkML class slot
    """

    name: SAFE_NAME
    range: TYPE_EXPRESSION = None
    default_value: str = None
    annotations: List[ANNOTATION] = field(default_factory=lambda: [])
    source_slot: SlotDefinition = field(default_factory=lambda: [])


@dataclass
class OOClass:
    """
    An object-oriented class
    """

    # ObjectVars
    name: SAFE_NAME
    description: Optional[SAFE_NAME] = None
    is_a: Optional[SAFE_NAME] = None
    mixin: Optional[bool] = None
    abstract: Optional[bool] = None
    mixins: List[SAFE_NAME] = field(default_factory=lambda: [])
    fields: List[OOField] = field(default_factory=lambda: [])
    all_fields: List[OOField] = field(default_factory=lambda: [])
    annotations: List[ANNOTATION] = field(default_factory=lambda: [])
    package: PACKAGE = None
    source_class: ClassDefinition = None


@dataclass
class OOCodeGenerator(Generator):

    # ClassVars
    java_style = True
    visit_all_class_slots = False
    uses_schemaloader = False
    requires_metamodel = False

    template_file: str = None
    """Path to template"""

    package: PACKAGE = "example"

    def __post_init__(self):
        # TODO: consider moving up a level
        self.schemaview = SchemaView(self.schema)
        super().__post_init__()

    @abc.abstractmethod
    def serialize(self, directory: str) -> None:
        raise NotImplementedError("Not implemented.")

    @abc.abstractmethod
    def default_value_for_type(self, typ: str) -> str:
        raise NotImplementedError

    def get_class_name(self, cn):
        return camelcase(cn)

    def get_slot_name(self, sn):
        if self.java_style:
            safe_sn = lcamelcase(sn)
        else:
            safe_sn = underscore(sn)
        return safe_sn

    def map_type(self, t: TypeDefinition, required: bool = False) -> str:
        return t.base

    def make_multivalued(self, range: str) -> str:
        return f"List<{range}>"

    def replace_invalid_identifier_character(self, char: str) -> str:
        if char.isalpha() or char.isnumeric() or char == "_":
            return char
        else:
            return underscore(unicodedata.name(char))

    def generate_enum_label(self, value: str) -> str:
        label = underscore(value)
        if label.isidentifier():
            return label
        else:
            # add an underscore if the value starts with a digit
            label = re.sub("(?=^\d)", "number_", label)

            safe_label = ""
            for character in label:
                safe_label += self.replace_invalid_identifier_character(character)

            return safe_label

    def create_documents(self) -> List[OODocument]:
        """
        Currently hardcoded for java-style
        :return:
        """
        sv: SchemaView
        sv = self.schemaview
        docs = []
        for cn in sv.all_classes(imports=False):
            c = sv.get_class(cn)
            safe_cn = camelcase(cn)
            oodoc = OODocument(
                name=safe_cn, package=self.package, source_schema=sv.schema
            )
            docs.append(oodoc)
            ooclass = OOClass(
                name=safe_cn, description=c.description, package=self.package, fields=[], source_class=c
            )
            # currently hardcoded for java style, one class per doc
            oodoc.classes = [ooclass]
            if c.mixin:
                ooclass.mixin = c.mixin
            if c.mixins:
                ooclass.mixins = [camelcase(x) for x in c.mixins]
            if c.abstract:
                ooclass.abstract = c.abstract
            if c.is_a:
                ooclass.is_a = self.get_class_name(c.is_a)
                parent_slots = sv.class_slots(c.is_a)
            else:
                parent_slots = []
            for sn in sv.class_slots(cn):
                safe_sn = self.get_slot_name(sn)
                slot = sv.induced_slot(sn, cn)
                range = slot.range
                default_value = "null"

                if range is None:
                    # TODO: schemaview should infer this
                    range = sv.schema.default_range

                if range is None:
                    range = "string"

                if range in sv.all_classes():
                    range = self.get_class_name(range)
                    default_value = "null"
                elif range in sv.all_types():
                    t = sv.get_type(range)
                    range = self.map_type(t, slot.required)
                    if range is None:  # If mapping fails,
                        range = self.map_type(sv.all_types().get("string"))
                elif range in sv.all_enums():
                    range = self.map_type(sv.all_types().get("string"))
                else:
                    raise Exception(f"Unknown range {range}")

                # Set default values for
                if range == "boolean":
                    default_value = "false"
                elif range == "integer":
                    default_value = "0"
                elif range == "String":
                    default_value = '""'

                # TODO:
                #  default_value = default_value_for_type(range)

                if slot.multivalued:
                    range = self.make_multivalued(range)
                    default_value = "List.of()"
                oofield = OOField(
                    name=safe_sn,
                    source_slot=slot,
                    range=range,
                    default_value=default_value,
                )
                if sn not in parent_slots:
                    ooclass.fields.append(oofield)
                ooclass.all_fields.append(oofield)

        return docs



import copy
from datetime import datetime
from typing import Optional, TextIO, Union
from urllib.parse import urlparse

import yaml
from dateutil.parser import ParserError, parse
from hbreader import FileInfo, HBType, detect_type
from linkml_runtime.linkml_model.meta import (ClassDefinition,
                                              SchemaDefinition, SlotDefinition,
                                              metamodel_version)
from linkml_runtime.loaders import yaml_loader
from linkml_runtime.utils.yamlutils import YAMLMark, YAMLRoot

from linkml.utils.mergeutils import set_from_schema

DATETIME_FORMAT = "%Y-%m-%dT%H:%M:%S"
yaml.error.Mark = YAMLMark


# Override the default linkml missing value tests
def mrf(self, field_name: str) -> None:
    if isinstance(self, SchemaDefinition) and field_name == "name" and self.id:
        id_parts = self.id.replace("#", "/").rsplit("/")
        self.name = id_parts[-1]
    else:
        YAMLRoot.MissingRequiredField(self, f"{type(self).__name__}.{field_name}")


SchemaDefinition.MissingRequiredField = mrf


def load_raw_schema(
    data: Union[str, dict, TextIO],
    source_file: Optional[str] = None,
    source_file_date: Optional[str] = None,
    source_file_size: Optional[int] = None,
    base_dir: Optional[str] = None,
    merge_modules: Optional[bool] = True,
    emit_metadata: Optional[bool] = True,
) -> SchemaDefinition:
    """Load and flatten SchemaDefinition from a file name, a URL or a block of text

    @param data: URL, file name or block of text YAML Object or open file handle
    @param source_file: Source file name for the schema if data is type TextIO
    @param source_file_date: timestamp of source file if data is type TextIO
    @param source_file_size: size of source file if data is type TextIO
    @param base_dir: Working directory or base URL of sources
    @param merge_modules: True means combine modules into one source, false means keep separate
    @param emit_metadata: True means add source file info to the output
    @return: Un-processed Schema Definition object
    """

    def _name_from_url(url) -> str:
        return urlparse(url).path.rsplit("/", 1)[-1].rsplit(".", 1)[0]

    # Passing a URL or file name
    if detect_type(data, base_dir) not in (HBType.STRING, HBType.STRINGABLE):
        assert (
            source_file is None
        ), "source_file parameter not allowed if data is a file or URL"
        assert (
            source_file_date is None
        ), "source_file_date parameter not allowed if data is a file or URL"
        assert (
            source_file_size is None
        ), "source_file_size parameter not allowed if data is a file or URL"

    # Convert the input into a valid SchemaDefinition
    if isinstance(data, (str, dict, TextIO)):
        # TODO: Build a generic loader that detects type from suffix or content and invokes the appropriate loader
        schema_metadata = FileInfo()
        schema_metadata.source_file = source_file
        schema_metadata.source_file_date = source_file_date
        schema_metadata.source_file_size = source_file_size
        schema_metadata.base_path = base_dir
        schema = yaml_loader.load(
            copy.deepcopy(data) if isinstance(data, dict) else data,
            SchemaDefinition,
            base_dir=base_dir,
            metadata=schema_metadata,
        )
    elif isinstance(data, SchemaDefinition):
        schema = copy.deepcopy(data)
    else:
        raise ValueError("Unrecognized input to raw loader")

    if schema is None:
        raise ValueError("Empty schema - cannot process")

    if schema.name is None:
        if schema.id is None:
            raise ValueError("Unable to determine schema name")
        else:
            schema.name = _name_from_url(schema.id)
    elif schema.id is None:
        # TODO: figure out how to generate this from the default_prefix and namespace map
        raise ValueError("Schema identifier must be supplied")

    if emit_metadata:
        schema.source_file = schema_metadata.source_file
        src_date = schema_metadata.source_file_date
        try:
            schema.source_file_date = (
                parse(src_date).strftime(DATETIME_FORMAT) if src_date else None
            )
        except ParserError:
            schema.source_file_date = src_date
        schema.source_file_size = schema_metadata.source_file_size
        schema.generation_date = datetime.now().strftime(DATETIME_FORMAT)
    schema.metamodel_version = metamodel_version

    set_from_schema(schema)

    return schema



import logging
import os
from collections import defaultdict
from dataclasses import dataclass, field
from functools import lru_cache
from pathlib import Path
from typing import Any, Dict, List, Tuple, Type, Union

import click
import yaml

from linkml._version import __version__
from linkml.generators.excelgen import ExcelGenerator
from linkml.generators.graphqlgen import GraphqlGenerator
from linkml.generators.javagen import JavaGenerator
from linkml.generators.jsonldcontextgen import ContextGenerator
from linkml.generators.jsonldgen import JSONLDGenerator
from linkml.generators.jsonschemagen import JsonSchemaGenerator
from linkml.generators.markdowngen import MarkdownGenerator
from linkml.generators.owlgen import OwlSchemaGenerator
from linkml.generators.prefixmapgen import PrefixGenerator
from linkml.generators.protogen import ProtoGenerator
from linkml.generators.pythongen import PythonGenerator
from linkml.generators.shaclgen import ShaclGenerator
from linkml.generators.shexgen import ShExGenerator
from linkml.generators.sqlddlgen import SQLDDLGenerator
from linkml.utils.generator import Generator, shared_arguments

PATH_FSTRING = str
GENERATOR_NAME = str
ARG_DICT = Dict[str, Any]
CONFIG_TUPLE = Tuple[Type[Generator], PATH_FSTRING, ARG_DICT]
GEN_MAP: Dict[GENERATOR_NAME, CONFIG_TUPLE]
GEN_MAP = {
    "graphql": (GraphqlGenerator, "graphql/{name}.graphql", {}),
    "jsonldcontext": (ContextGenerator, "jsonld/{name}.context.jsonld", {}),
    "jsonld": (
        JSONLDGenerator,
        "jsonld/{name}.jsonld",
        {"context": "{parent}/{name}.context.jsonld"},
    ),
    "jsonschema": (JsonSchemaGenerator, "jsonschema/{name}.schema.json", {}),
    "markdown": (
        MarkdownGenerator,
        "docs/",
        {"directory": "{parent}", "index_file": "{name}.md"},
    ),
    "owl": (OwlSchemaGenerator, "owl/{name}.owl.ttl", {}),
    "prefixmap": (PrefixGenerator, "prefixmap/{name}.yaml", {}),
    "proto": (ProtoGenerator, "protobuf/{name}.proto", {}),
    "python": (PythonGenerator, "{name}.py", {}),
    #    'rdf': (RDFGenerator, 'rdf/{name}.ttl', {}),
    #    'rdf': (RDFGenerator, 'rdf/{name}.ttl', {'context': '{parent}/../jsonld/{name}.context.jsonld'}),
    "shex": (ShExGenerator, "shex/{name}.shex", {}),
    "shacl": (ShaclGenerator, "shacl/{name}.shacl.ttl", {}),
    "sqlddl": (SQLDDLGenerator, "sqlschema/{name}.sql", {}),
    # # linkml/generators/javagen.py uses different architecture from most of the other generators
    # # also linkml/generators/excelgen.py, which has a different mechanism for determining the output path
    # 'java': (JavaGenerator, 'java/{name}.java', {'directory': '{parent}'}),
    "excel": (ExcelGenerator, "excel/{name}.xlsx", {"output": "{parent}/{name}.xlsx"}),
}


@lru_cache()
def get_local_imports(schema_path: str, dir: str):
    logging.info(f"GETTING IMPORTS = {schema_path}")
    all_imports = [schema_path]
    with open(schema_path) as stream:
        with open(schema_path) as stream:
            schema = yaml.safe_load(stream)
            for imp in schema.get("imports", []):
                imp_path = os.path.join(dir, imp) + ".yaml"
                logging.info(f" IMP={imp} //  path={imp_path}")
                if os.path.isfile(imp_path):
                    all_imports += get_local_imports(imp_path, dir)
    return all_imports


@dataclass
class ProjectConfiguration:
    """
    Global project configuration, and per-generator configurations
    """

    directory: str = "tmp"
    generator_args: Dict[GENERATOR_NAME, ARG_DICT] = field(
        default_factory=lambda: defaultdict(dict)
    )
    includes: List[str] = None
    excludes: List[str] = None
    mergeimports: bool = None


class ProjectGenerator:
    """
    Generates complete project folders

    Note this doesn't conform to overall generator framework, as it is a meta-generator
    """

    def generate(
        self, schema_path: str, config: ProjectConfiguration = ProjectConfiguration()
    ):
        if config.directory is None:
            raise Exception(f"Must pass directory")
        Path(config.directory).mkdir(parents=True, exist_ok=True)
        if config.mergeimports:
            all_schemas = [schema_path]
        else:
            all_schemas = get_local_imports(schema_path, os.path.dirname(schema_path))
        print(f"ALL_SCHEMAS = {all_schemas}")
        for gen_name, (gen_cls, gen_path_fmt, default_gen_args) in GEN_MAP.items():
            if (
                config.includes is not None
                and config.includes != []
                and gen_name not in config.includes
            ):
                logging.info(
                    f"Skipping {gen_name} as not in inclusion list: {config.includes}"
                )
                continue
            if config.excludes is not None and gen_name in config.excludes:
                logging.info(f"Skipping {gen_name} as it is in exclusion list")
                continue
            logging.info(f"Generating: {gen_name}")
            for local_path in all_schemas:
                logging.info(f" SCHEMA: {local_path}")
                name = os.path.basename(local_path).replace(".yaml", "")
                gen_path = gen_path_fmt.format(name=name)
                gen_path_full = f"{config.directory}/{gen_path}"
                parts = gen_path_full.split("/")
                parent_dir = "/".join(parts[0:-1])
                logging.info(f" PARENT={parent_dir}")
                Path(parent_dir).mkdir(parents=True, exist_ok=True)
                gen_path_full = "/".join(parts)
                all_gen_args = {
                    **default_gen_args,
                    **config.generator_args.get(gen_name, {}),
                }
                gen: Generator

                # special check for output key because ExcelGenerator and
                # SSSOMGenerator read in output file name during initialization
                if "output" in all_gen_args:
                    all_gen_args["output"] = all_gen_args["output"].format(
                        name=name, parent=parent_dir
                    )

                gen = gen_cls(local_path, **all_gen_args)

                serialize_args = {"mergeimports": config.mergeimports}
                for k, v in all_gen_args.items():
                    # all ARG_DICT values are interpolatable
                    if isinstance(v, str):
                        v = v.format(name=name, parent=parent_dir)
                    serialize_args[k] = v
                logging.info(f" {gen_name} ARGS: {serialize_args}")
                gen_dump = gen.serialize(**serialize_args)

                if gen_name != "excel":
                    if parts[-1] != "":
                        # markdowngen does not write to a file
                        logging.info(f"  WRITING TO: {gen_path_full}")
                        with open(gen_path_full, "w", encoding="UTF-8") as stream:
                            stream.write(gen_dump)
                else:
                    # special handling for excel generator
                    # we do not need to route the output
                    # into a file like the other generators
                    gen.serialize(**serialize_args)


@click.command()
@click.option(
    "--dir",
    "-d",
    help="directory in which to place generated files. E.g. linkml_model, biolink_model",
)
@click.option("--generator-arguments", "-A", help="yaml configuration for generators")
@click.option(
    "--config-file", "-C", type=click.File("rb"), help="path to yaml configuration"
)
@click.option(
    "--exclude", "-X", multiple=True, help="list of artefacts to be excluded"
)  # TODO: make this an enum
@click.option(
    "--include",
    "-I",
    multiple=True,
    help="list of artefacts to be included. If not set, defaults to all",
)  # TODO: make this an enum
@click.option(
    "--mergeimports/--no-mergeimports",
    default=True,
    show_default=True,
    help="Merge imports into source file",
)
@click.argument("yamlfile")
@click.version_option(__version__, "-V", "--version")
def cli(
    yamlfile,
    dir,
    exclude: List[str],
    include: List[str],
    config_file,
    mergeimports,
    generator_arguments: str,
    **kwargs,
):
    """
    Generate an entire project LinkML schema

    Generate all downstream artefacts using default configuration:

       gen-project -d . personinfo.yaml

    Exclusion lists: all except ShEx:

       gen-project --exclude shex -d . personinfo.yaml

    Inclusion lists: only jsonschema and python:

       gen-project -I python -I jsonschema -d . personinfo.yaml

    Configuration, on command line:

        gen-project -A 'jsonschema: {top_class: Container}' -d . personinfo.yaml

    Configuration, via yaml file:

        gen-project --config config.yaml personinfo.yaml

    config.yaml:

        directory: .
        generator_args:
          json_schema:
            top_class: Container

    """
    logging.basicConfig(level=logging.INFO)
    project_config = ProjectConfiguration()
    if config_file is not None:
        for k, v in yaml.safe_load(config_file).items():
            setattr(project_config, k, v)
    if exclude:
        project_config.excludes = list(exclude)
    if include:
        project_config.includes = list(include)
    if generator_arguments is not None:
        try:
            project_config.generator_args = yaml.safe_load(generator_arguments)
        except Exception:
            raise Exception(f"Argument must be a valid YAML blob")
        logging.info(f"generator args: {project_config.generator_args}")
    if dir is not None:
        project_config.directory = dir
    project_config.mergeimports = mergeimports
    gen = ProjectGenerator()
    gen.generate(yamlfile, project_config)


if __name__ == "__main__":
    cli()



"""
Generate JSON-LD contexts

"""
import os
import re
from dataclasses import dataclass, field
from typing import Optional, Set, TextIO, Union, Dict

import click
from jsonasobj2 import JsonObj, as_json
from linkml_runtime.linkml_model.meta import (ClassDefinition, Definition,
                                              Element, SchemaDefinition,
                                              SlotDefinition)
from linkml_runtime.linkml_model.types import SHEX
from linkml_runtime.utils.formatutils import be, camelcase, underscore
from rdflib import SKOS, XSD

from linkml._version import __version__
from linkml.utils.generator import Generator, shared_arguments

URI_RANGES = (XSD.anyURI, SHEX.nonliteral, SHEX.bnode, SHEX.iri)


ENUM_CONTEXT = {
    "@vocab": "@null",
    "text": "skos:notation",
    "description": "skos:prefLabel",
    "meaning": "@id",
}


@dataclass
class ContextGenerator(Generator):

    # ClassVars
    generatorname = os.path.basename(__file__)
    generatorversion = "0.1.1"
    valid_formats = ["context", "json"]
    visit_all_class_slots = False
    uses_schemaloader = True
    requires_metamodel = True

    # ObjectVars
    emit_prefixes: Set[str] = field(default_factory=lambda: set())
    default_ns: str = None
    context_body: Dict = field(default_factory=lambda: dict())
    slot_class_maps: Dict = field(default_factory=lambda: dict())
    emit_metadata: bool = field(default_factory=lambda: False)
    model: Optional[bool] = field(default_factory=lambda: True)
    base: Optional[str] = None
    output: Optional[str] = None
    prefixes: Optional[bool] = field(default_factory=lambda: True)
    flatprefixes: Optional[bool] = field(default_factory=lambda: False)

    def __post_init__(self) -> None:
        super().__post_init__()
        if self.namespaces is None:
            raise TypeError(
                "Schema text must be supplied to context generator.  Preparsed schema will not work"
            )

    def visit_schema(
        self, base: Optional[str] = None, output: Optional[str] = None, **_
    ):
        # Add any explicitly declared prefixes
        for prefix in self.schema.prefixes.values():
            self.emit_prefixes.add(prefix.prefix_prefix)

        # Add any prefixes explicitly declared
        for pfx in self.schema.emit_prefixes:
            self.add_prefix(pfx)

        # Add the default prefix
        if self.schema.default_prefix:
            dflt = self.namespaces.prefix_for(self.schema.default_prefix)
            if dflt:
                self.default_ns = dflt
            if self.default_ns:
                default_uri = self.namespaces[self.default_ns]
                self.emit_prefixes.add(self.default_ns)
            else:
                default_uri = self.schema.default_prefix
                if self.schema.name:
                    self.namespaces[self.schema.name] = default_uri
                    self.emit_prefixes.add(self.schema.name)
            self.context_body["@vocab"] = default_uri
            # self.context_body['@base'] = self.base_dir

    def end_schema(
        self,
        base: Optional[str] = None,
        output: Optional[str] = None,
        prefixes: Optional[bool] = True,
        flatprefixes: Optional[bool] = False,
        model: Optional[bool] = True,
        **_,
    ) -> None:
        if model is None:
            model = self.model
        context = JsonObj()
        if self.emit_metadata:
            comments = JsonObj()
            comments.description = "Auto generated by LinkML jsonld context generator"
            comments.generation_date = self.schema.generation_date
            comments.source = self.schema.source_file
    #
    #         comments = f"""Auto generated from {self.schema.source_file} by {self.generatorname} version: {self.generatorversion}"""
    #         if self.schema.generation_date:
    #             comments += f"""
    #
    # Schema: {self.schema.name}
    # metamodel version: {self.schema.metamodel_version}
    # model version: {self.schema.version if self.schema.version else None}
    # """
    #         comments += f"""
    # id: {self.schema.id}
    # description: {be(self.schema.description)}
    # license: {be(self.schema.license)}
    # """
            context.comments = comments
        context_content = {}
        if base:
            if "://" not in base:
                self.context_body["@base"] = os.path.relpath(
                    base, os.path.dirname(self.schema.source_file)
                )
            else:
                self.context_body["@base"] = base
        if prefixes:
            for prefix in sorted(self.emit_prefixes):
                url = str(self.namespaces[prefix])
                # Derived from line # ~5223 in pyld/lib/jsonld.py
                if bool(re.match(r".*[:/\?#\[\]@]$", url)) or flatprefixes:
                    context_content[prefix] = url
                else:
                    prefix_obj = JsonObj()
                    prefix_obj["@id"] = url
                    prefix_obj["@prefix"] = True
                    context_content[prefix] = prefix_obj
        if model:
            for k, v in self.context_body.items():
                context_content[k] = v
            for k, v in self.slot_class_maps.items():
                context_content[k] = v
        context["@context"] = context_content
        if output:
            with open(output, "w", encoding="UTF-8") as outf:
                outf.write(as_json(context))
        else:
            print(as_json(context))

    def visit_class(self, cls: ClassDefinition) -> bool:
        class_def = {}
        cn = camelcase(cls.name)
        self.add_mappings(cls)
        cls_uri_prefix, cls_uri_suffix = self.namespaces.prefix_suffix(cls.class_uri)
        if (
            not self.default_ns
            or not cls_uri_prefix
            or cls_uri_prefix != self.default_ns
        ):
            class_def["@id"] = (
                (cls_uri_prefix + ":" + cls_uri_suffix)
                if cls_uri_prefix
                else cls.class_uri
            )
            if cls_uri_prefix:
                self.add_prefix(cls_uri_prefix)
        if class_def:
            self.slot_class_maps[cn] = class_def

        # We don't bother to visit class slots - just all slots
        return False

    def visit_slot(self, aliased_slot_name: str, slot: SlotDefinition) -> None:
        if slot.identifier:
            slot_def = "@id"
        else:
            slot_def = {}
            if not slot.usage_slot_name:
                if slot.range in self.schema.classes:
                    slot_def["@type"] = "@id"
                elif slot.range in self.schema.enums:
                    slot_def["@context"] = ENUM_CONTEXT
                    # Add the necessary prefixes to the namespace
                    skos = self.namespaces.prefix_for(SKOS)
                    if not skos:
                        self.namespaces["skos"] = SKOS
                        skos = "skos"
                    self.emit_prefixes.add(skos)
                else:
                    range_type = self.schema.types[slot.range]
                    if self.namespaces.uri_for(range_type.uri) == XSD.string:
                        pass
                    elif self.namespaces.uri_for(range_type.uri) in URI_RANGES:
                        slot_def["@type"] = "@id"
                    else:
                        slot_def["@type"] = range_type.uri
                slot_prefix = self.namespaces.prefix_for(slot.slot_uri)
                if (
                    not self.default_ns
                    or not slot_prefix
                    or slot_prefix != self.default_ns
                ):
                    slot_def["@id"] = slot.slot_uri
                    if slot_prefix:
                        self.add_prefix(slot_prefix)
                self.add_mappings(slot)
        if slot_def:
            self.context_body[underscore(aliased_slot_name)] = slot_def


@shared_arguments(ContextGenerator)
@click.command()
@click.option("--base", help="Base URI for model")
@click.option(
    "--prefixes/--no-prefixes",
    default=True,
    show_default=True,
    help="Emit context for prefixes (default=--prefixes)",
)
@click.option(
    "--model/--no-model",
    default=True,
    show_default=True,
    help="Emit context for model elements (default=--model)",
)
@click.option(
    "--flatprefixes/--no-flatprefixes",
    default=False,
    show_default=True,
    help="Emit non-JSON-LD compliant prefixes as an object (deprecated: use gen-prefix-map instead).",
)
@click.version_option(__version__, "-V", "--version")
def cli(yamlfile, **args):
    """Generate jsonld @context definition from LinkML model"""
    print(ContextGenerator(yamlfile, **args).serialize(**args))


if __name__ == "__main__":
    cli()



import inspect
from copy import deepcopy
from dataclasses import dataclass
from functools import lru_cache
from pathlib import Path
from typing import Any, Dict, Iterable, Union

import yaml
from linkml_runtime import SchemaView
from linkml_runtime.dumpers import yaml_dumper
from linkml_runtime.linkml_model import SchemaDefinition

from .config.datamodel.config import Config, ExtendableConfigs, RuleLevel


@dataclass
class LinterProblem:
    message: str
    level: Union[RuleLevel, None] = None
    schema_name: Union[str, None] = None
    schema_source: Union[str, None] = None
    rule_name: Union[str, None] = None


@lru_cache
def get_named_config(name: str) -> Dict[str, Any]:
    config_path = str(Path(__file__).parent / f"config/{name}.yaml")
    with open(config_path) as config_file:
        return yaml.safe_load(config_file)


def merge_configs(original: dict, other: dict):
    result = deepcopy(original)
    for key, value in other.items():
        if isinstance(value, dict):
            result[key] = merge_configs(result.get(key, {}), value)
        else:
            result[key] = value
    return result


class Linter:
    def __init__(self, config: Dict[str, Any] = {}) -> None:
        default_config = deepcopy(get_named_config("default"))
        merged_config = config
        if config.get("extends") == ExtendableConfigs.recommended.text:
            recommended_config = deepcopy(
                get_named_config(ExtendableConfigs.recommended.text)
            )
            merged_config = merge_configs(recommended_config, merged_config)
        merged_config = merge_configs(default_config, merged_config)
        self.config = Config(**merged_config)

        from . import rules

        self._rules_map = dict(
            [
                (cls.id, cls)
                for _, cls in inspect.getmembers(rules, inspect.isclass)
                if issubclass(cls, rules.LinterRule)
            ]
        )

    def lint(
        self, schema=Union[str, SchemaDefinition], fix: bool = False
    ) -> Iterable[LinterProblem]:
        try:
            schema_view = SchemaView(schema)
        except:
            yield LinterProblem(
                message="File is not a valid LinkML schema",
                level=RuleLevel(RuleLevel.error),
                schema_source=(schema if isinstance(schema, str) else None),
            )
            return

        for rule_id, rule_config in self.config.rules.__dict__.items():
            rule_cls = self._rules_map.get(rule_id, None)
            if rule_cls is None:
                raise ValueError("Unknown rule id: " + rule_id)

            if str(rule_config.level) is RuleLevel.disabled.text:
                continue

            rule = rule_cls(rule_config)

            for problem in rule.check(schema_view, fix=fix):
                problem.level = rule.config.level
                problem.rule_name = rule.id
                problem.schema_name = schema_view.schema.name
                if isinstance(schema, str):
                    problem.schema_source = schema
                yield problem

        if fix and schema_view.schema.source_file:
            yaml_dumper.dump(schema_view.schema, schema_view.schema.source_file)



import keyword
import logging
import os
import re
from dataclasses import dataclass, field
from types import ModuleType
from typing import (Callable, Dict, Iterator, List, Optional, Set, TextIO,
                    Tuple, Union)

import click
from linkml_runtime.linkml_model import linkml_files
from linkml_runtime.linkml_model.meta import (ClassDefinition,
                                              ClassDefinitionName, Definition,
                                              DefinitionName, Element,
                                              EnumDefinition, PermissibleValue,
                                              SchemaDefinition, SlotDefinition,
                                              SlotDefinitionName,
                                              TypeDefinition)
from linkml_runtime.utils.compile_python import compile_python
from linkml_runtime.utils.formatutils import (be, camelcase, sfx, split_line,
                                              underscore, wrapped_annotation)
from linkml_runtime.utils.metamodelcore import builtinnames
from rdflib import URIRef

import linkml
from linkml._version import __version__
from linkml.generators import PYTHON_GEN_VERSION
from linkml.utils.generator import Generator, shared_arguments
from linkml.utils.ifabsent_functions import (default_curie_or_uri,
                                             ifabsent_postinit_declaration,
                                             ifabsent_value_declaration)


@dataclass
class PythonGenerator(Generator):
    """
    Generates Python dataclasses from a LinkML model

    See `Python Generator Docs <https://linkml.io/linkml/generators/python.html>`_
    """

    # ClassVars
    generatorname = os.path.basename(__file__)
    generatorversion = PYTHON_GEN_VERSION
    valid_formats = ["py"]
    visit_all_class_slots = False
    uses_schemaloader = True

    # ObjectVars
    gen_classvars: bool = field(default_factory=lambda: True)
    gen_slots: bool = field(default_factory=lambda: True)
    genmeta: bool = field(default_factory=lambda: False)
    emit_metadata: bool = field(default_factory=lambda: True)

    def __post_init__(self) -> None:
        self.sourcefile = self.schema
        super().__post_init__()
        if self.format is None:
            self.format = self.valid_formats[0]
        if self.schema.default_prefix == "linkml" and not self.genmeta:
            logging.error(
                f"Generating metamodel without --genmeta is highly inadvised!"
            )
        if (
            not self.schema.source_file
            and isinstance(self.sourcefile, str)
            and "\n" not in self.sourcefile
        ):
            self.schema.source_file = os.path.basename(self.sourcefile)

    def compile_module(self, **kwargs) -> ModuleType:
        """
        Compiles generated python code to a module
        :return:
        """
        pycode = self.serialize(**kwargs)
        try:
            return compile_python(pycode)
        except NameError as e:
            logging.error(f"Code:\n{pycode}")
            logging.error(f"Error compiling generated python code: {e}")
            raise e

    def visit_schema(self, **kwargs) -> None:
        # Add explicitly declared prefixes
        self.emit_prefixes.update(
            [p.prefix_prefix for p in self.schema.prefixes.values()]
        )

        # Add all emit statements
        self.emit_prefixes.update(self.schema.emit_prefixes)

        # Add the default prefix
        if self.schema.default_prefix:
            self.emit_prefixes.add(
                self.namespaces.prefix_for(self.schema.default_prefix)
            )

    def visit_class(self, cls: ClassDefinition) -> bool:
        if not cls.imported_from:
            cls_prefix = self.namespaces.prefix_for(cls.class_uri)
            if cls_prefix:
                self.emit_prefixes.add(cls_prefix)
            self.add_mappings(cls)
        return False

    def visit_slot(self, aliased_slot_name: str, slot: SlotDefinition) -> None:
        if not slot.imported_from:
            slot_prefix = self.namespaces.prefix_for(slot.slot_uri)
            if slot_prefix:
                self.emit_prefixes.add(slot_prefix)
            self.add_mappings(slot)

    def visit_type(self, typ: TypeDefinition) -> None:
        if not typ.imported_from:
            type_prefix = self.namespaces.prefix_for(typ.uri)
            if type_prefix:
                self.emit_prefixes.add(type_prefix)

    def gen_schema(self) -> str:
        # The metamodel uses Enumerations to define itself, so don't import if we are generating the metamodel
        enumimports = (
            ""
            if self.genmeta
            else "from linkml_runtime.linkml_model.meta import EnumDefinition, PermissibleValue, PvFormulaOptions\n"
        )
        handlerimport = (
            "from linkml_runtime.utils.enumerations import EnumDefinitionImpl"
        )
        split_descripton = "\n#              ".join(
            split_line(be(self.schema.description), split_len=100)
        )
        head = (
            f"""# Auto generated from {self.schema.source_file} by {self.generatorname} version: {self.generatorversion}
# Generation date: {self.schema.generation_date}
# Schema: {self.schema.name}
#"""
            if self.emit_metadata and self.schema.generation_date
            else ""
        )

        return f"""{head}
# id: {self.schema.id}
# description: {split_descripton}
# license: {be(self.schema.license)}

import dataclasses
import sys
import re
from jsonasobj2 import JsonObj, as_dict
from typing import Optional, List, Union, Dict, ClassVar, Any
from dataclasses import dataclass
{enumimports}
from linkml_runtime.utils.slot import Slot
from linkml_runtime.utils.metamodelcore import empty_list, empty_dict, bnode
from linkml_runtime.utils.yamlutils import YAMLRoot, extended_str, extended_float, extended_int
from linkml_runtime.utils.dataclass_extensions_376 import dataclasses_init_fn_with_kwargs
from linkml_runtime.utils.formatutils import camelcase, underscore, sfx
{handlerimport}
from rdflib import Namespace, URIRef
from linkml_runtime.utils.curienamespace import CurieNamespace
{self.gen_imports()}

metamodel_version = "{self.schema.metamodel_version}"
version = {'"' + self.schema.version + '"' if self.schema.version else None}

# Overwrite dataclasses _init_fn to add **kwargs in __init__
dataclasses._init_fn = dataclasses_init_fn_with_kwargs

# Namespaces
{self.gen_namespaces()}


# Types
{self.gen_typedefs()}
# Class references
{self.gen_references()}

{self.gen_classdefs()}

# Enumerations
{self.gen_enumerations()}

# Slots
{self.gen_slotdefs()}"""

    def end_schema(self, **_):
        print(
            re.sub(r" +\n", "\n", self.gen_schema().replace("\t", "    ")).strip(" "),
            end="",
        )

    def gen_imports(self) -> str:
        listents = [
            f"from {k} import {', '.join(v)}" for k, v in self.gen_import_list().items()
        ]
        return "\n".join(listents)

    def gen_import_list(self) -> Dict[str, List[str]]:
        """
        Generate a list of types to import

        :return: source file followed by elements to import
        """

        class ImportList:
            def __init__(self, schema_location: str):
                self.schema_location = schema_location
                self.v: Dict[str, Set[str]] = {}

            def add_element(self, e: Element) -> None:
                if e.imported_from:
                    self.add_entry(e.imported_from, camelcase(e.name))

            def add_entry(innerself, path: Union[str, URIRef], name: str) -> None:
                path = str(self.namespaces.uri_for(path) if ":" in path else path)
                if path.startswith(linkml_files.LINKML_NAMESPACE):
                    model_base = "." if self.genmeta else "linkml_runtime.linkml_model."
                    innerself.v.setdefault(
                        model_base + path[len(linkml_files.LINKML_NAMESPACE) :], set()
                    ).add(name)
                elif path == linkml.BIOLINK_MODEL_URI:
                    innerself.v.setdefault(linkml.BIOLINK_MODEL_PYTHON_LOC, set()).add(
                        name
                    )
                elif "://" in path:
                    raise ValueError(
                        f"Cannot map {path} into a python import statement"
                    )
                elif "/" in path:
                    innerself.v.setdefault(
                        path.replace("./", ".").replace("/", "."), set()
                    ).add(name)
                elif "." in path:
                    innerself.v.setdefault(path, set()).add(name)
                else:
                    innerself.v.setdefault(". " + path, set()).add(name)

            def values(self) -> Dict[str, List[str]]:
                return {k: sorted(self.v[k]) for k in sorted(self.v.keys())}

        def add_type_ref(typ: TypeDefinition) -> None:
            if not typ.typeof and typ.base and typ.base not in builtinnames:
                if "." in typ.base:
                    rval.add_entry(*typ.base.rsplit("."))
                else:
                    rval.add_entry("linkml_runtime.utils.metamodelcore", typ.base)
            if typ.typeof:
                add_type_ref(self.schema.types[typ.typeof])
            rval.add_element(typ)

        def add_enum_ref(e: EnumDefinition) -> None:
            rval.add_element(e)

        def add_slot_range(slot: SlotDefinition) -> None:
            if slot.range:
                if slot.range in self.schema.types:
                    add_type_ref(self.schema.types[slot.range])
                elif slot.range in self.schema.enums:
                    add_enum_ref(self.schema.enums[slot.range])
                else:
                    cls = self.schema.classes[slot.range]
                    if cls.imported_from:
                        if self.class_identifier(cls):
                            identifier_range = self.class_identifier_path(cls, False)[
                                -1
                            ]
                            if identifier_range in self.schema.types:
                                add_type_ref(TypeDefinition(identifier_range))
                            else:
                                rval.add_entry(cls.imported_from, identifier_range)
                        if slot.inlined:
                            rval.add_element(cls)

        rval = ImportList(self.schema_location)
        for typ in self.schema.types.values():
            if not typ.imported_from:
                add_type_ref(typ)
        for slot in self.schema.slots.values():
            if not slot.imported_from:
                if slot.is_a:
                    parent = self.schema.slots[slot.is_a]
                    if (parent.key or parent.identifier) and parent.imported_from:
                        rval.add_element(self.schema.slots[slot.is_a])
                if slot.domain:
                    domain = self.schema.classes[slot.domain]
                    if domain.imported_from:
                        rval.add_element(self.schema.classes[slot.domain])
                add_slot_range(slot)

        for cls in self.schema.classes.values():
            if not cls.imported_from:
                if cls.is_a:
                    parent = self.schema.classes[cls.is_a]
                    if parent.imported_from:
                        rval.add_element(self.schema.classes[cls.is_a])
                        if self.class_identifier(parent):
                            rval.add_entry(
                                parent.imported_from,
                                self.class_identifier_path(parent, False)[-1],
                            )
                for slotname in cls.slots:
                    add_slot_range(self.schema.slots[slotname])
                # for slotname in cls.slot_usage:
                #     add_slot_range(self.schema.slots[slotname])

        return rval.values()

    def gen_namespaces(self) -> str:
        dflt_prefix = default_curie_or_uri(self)
        dflt = (
            f"CurieNamespace('', '{sfx(dflt_prefix)}')"
            if ":/" in dflt_prefix
            else dflt_prefix.upper()
        )
        return "\n".join(
            [
                f"{pfx.upper().replace('.', '_').replace('-', '_')} = CurieNamespace('{pfx.replace('.', '_')}', '{self.namespaces[pfx]}')"
                for pfx in sorted(self.emit_prefixes)
                if pfx in self.namespaces
            ]
            + [f"DEFAULT_ = {dflt}"]
        )

    def gen_references(self) -> str:
        """Generate python type declarations for all identifiers (primary keys)"""
        rval = []
        for cls in self._sort_classes(self.schema.classes.values()):
            if not cls.imported_from:
                pkeys = self.primary_keys_for(cls)
                if pkeys:
                    for pk in pkeys:
                        classname = camelcase(cls.name) + camelcase(
                            self.aliased_slot_name(pk)
                        )
                        # If we've got a parent slot and the range of the parent is the range of the child, the
                        # child slot is a subclass of the parent.  Otherwise, the child range has been overridden,
                        # so the inheritence chain has been broken
                        parent_pk = (
                            self.class_identifier(cls.is_a) if cls.is_a else None
                        )
                        parent_pk_slot = (
                            self.schema.slots[parent_pk] if parent_pk else None
                        )
                        pk_slot = self.schema.slots[pk]
                        if parent_pk_slot and (
                            parent_pk_slot.name == pk
                            or pk_slot.range == parent_pk_slot.range
                        ):
                            parents = self.class_identifier_path(cls.is_a, False)
                        else:
                            parents = self.slot_range_path(pk_slot)
                        parent_cls = (
                            "extended_" + parents[-1]
                            if parents[-1] in ["str", "float", "int"]
                            else parents[-1]
                        )
                        rval.append(f"class {classname}({parent_cls}):\n\tpass")
                        break  # We only do the first primary key
        return "\n\n\n".join(rval)

    def gen_typedefs(self) -> str:
        """Generate python type declarations for all defined types"""
        rval = []
        for typ in self.schema.types.values():
            if not typ.imported_from:
                typname = camelcase(typ.name)
                desc = f'\n\t""" {typ.description} """' if typ.description else ""

                if typ.typeof:
                    parent_typename = camelcase(typ.typeof)
                    rval.append(
                        f"class {typname}({parent_typename}):{desc}\n\t{self.gen_type_meta(typ)}\n\n"
                    )
                else:
                    base_base = typ.base.rsplit(".")[-1]
                    rval.append(
                        f"class {typname}({base_base}):{desc}\n\t{self.gen_type_meta(typ)}\n\n"
                    )
        return "\n".join(rval)

    def gen_classdefs(self) -> str:
        """Create class definitions for all non-mixin classes in the model
        Note that apply_to classes are transformed to mixins
        """
        clist = self._sort_classes(self.schema.classes.values())
        return "\n".join([self.gen_classdef(v) for v in clist if not v.imported_from])

    def gen_classdef(self, cls: ClassDefinition) -> str:
        """Generate python definition for class cls"""

        parentref = f'({self.formatted_element_name(cls.is_a, True) if cls.is_a else "YAMLRoot"})'
        slotdefs = self.gen_class_variables(cls)
        postinits = self.gen_postinits(cls)

        wrapped_description = (
            f'\n\t"""\n\t{wrapped_annotation(be(cls.description))}\n\t"""'
            if be(cls.description)
            else ""
        )

        if self.is_class_unconstrained(cls):
            return f"\n{self.class_or_type_name(cls.name)} = Any"

        return (
            ("\n@dataclass" if slotdefs else "")
            + f"\nclass {self.class_or_type_name(cls.name)}{parentref}:{wrapped_description}"
            + f"{self.gen_inherited_slots(cls)}"
            + f"{self.gen_class_meta(cls)}"
            + (f"\n\t{slotdefs}" if slotdefs else "")
            + (f"\n{postinits}" if postinits else "")
        )

    def gen_inherited_slots(self, cls: ClassDefinition) -> str:
        if not self.gen_classvars:
            return ""
        inherited_slots = []
        for slotname in cls.slots:
            slot = self.schema.slots[slotname]
            if slot.inherited:
                inherited_slots.append(slot.alias if slot.alias else slotname)
        inherited_slots_str = ", ".join([f'"{underscore(s)}"' for s in inherited_slots])
        return f"\n\t_inherited_slots: ClassVar[List[str]] = [{inherited_slots_str}]\n"

    def gen_class_meta(self, cls: ClassDefinition) -> str:
        if not self.gen_classvars:
            return ""
        class_class_uri = self.namespaces.uri_for(cls.class_uri)
        if class_class_uri:
            cls_python_uri = self.namespaces.curie_for(
                class_class_uri, default_ok=False, pythonform=True
            )
            class_class_curie = self.namespaces.curie_for(
                class_class_uri, default_ok=False, pythonform=False
            )
        else:
            cls_python_uri = None
            class_class_curie = None
        if class_class_curie:
            class_class_curie = f'"{class_class_curie}"'
        class_class_uri = (
            cls_python_uri if cls_python_uri else f'URIRef("{class_class_uri}")'
        )
        class_model_uri = self.namespaces.uri_or_curie_for(
            self.schema.default_prefix or "DEFAULT_", camelcase(cls.name)
        )
        if ":/" in class_model_uri:
            class_model_uri = f'URIRef("{class_model_uri}")'
        else:
            ns, ln = class_model_uri.split(":", 1)
            class_model_uri = f"{ns.upper()}.{ln}"

        vars = [
            f"class_class_uri: ClassVar[URIRef] = {class_class_uri}",
            f"class_class_curie: ClassVar[str] = {class_class_curie}",
            f'class_name: ClassVar[str] = "{cls.name}"',
            f"class_model_uri: ClassVar[URIRef] = {class_model_uri}",
        ]
        return "\n\t" + "\n\t".join(vars) + "\n"

    def gen_type_meta(self, typ: TypeDefinition) -> str:
        type_class_uri = self.namespaces.uri_for(typ.uri)
        if type_class_uri:
            type_python_uri = self.namespaces.curie_for(
                type_class_uri, default_ok=False, pythonform=True
            )
            type_class_curie = self.namespaces.curie_for(
                type_class_uri, default_ok=False, pythonform=False
            )
        else:
            type_python_uri = None
            type_class_curie = None
        if type_class_curie:
            type_class_curie = f'"{type_class_curie}"'
        type_class_uri = (
            type_python_uri if type_python_uri else f'URIRef("{type_class_uri}")'
        )
        type_model_uri = self.namespaces.uri_or_curie_for(
            self.schema.default_prefix, camelcase(typ.name)
        )
        if ":/" in type_model_uri:
            type_model_uri = f'URIRef("{type_model_uri}")'
        else:
            ns, ln = type_model_uri.split(":", 1)
            ln_suffix = f".{ln}" if ln.isidentifier() else f'["{ln}"]'
            type_model_uri = f"{ns.upper()}{ln_suffix}"
        vars = [
            f"type_class_uri = {type_class_uri}",
            f"type_class_curie = {type_class_curie}",
            f'type_name = "{typ.name}"',
            f"type_model_uri = {type_model_uri}",
        ]
        return "\n\t".join(vars)

    def gen_class_variables(self, cls: ClassDefinition) -> str:
        """
        Generate the variable declarations for a dataclass.

        :param cls: class containing variables to be rendered in inheritence hierarchy
        :return: variable declarations for target class and its ancestors
        """
        initializers = []

        is_root = not cls.is_a
        domain_slots = self.domain_slots(cls)

        # Root keys and identifiers go first.  Note that even if a key or identifier is overridden it still
        # appears at the top of the list, as we need to keep the position
        slot_variables = self._slot_iter(
            cls,
            lambda slot: (slot.identifier or slot.key) and not slot.ifabsent,
            first_hit_only=True,
        )
        initializers += [
            self.gen_class_variable(cls, slot, not is_root) for slot in slot_variables
        ]

        # Required slots
        slot_variables = self._slot_iter(
            cls,
            lambda slot: slot.required
            and not slot.identifier
            and not slot.key
            and not slot.ifabsent,
        )
        initializers += [
            self.gen_class_variable(cls, slot, not is_root) for slot in slot_variables
        ]

        # Required or key slots with default values
        slot_variables = self._slot_iter(
            cls, lambda slot: slot.ifabsent and slot.required
        )
        initializers += [
            self.gen_class_variable(cls, slot, False) for slot in slot_variables
        ]

        # Followed by everything else
        slot_variables = self._slot_iter(
            cls, lambda slot: not slot.required and slot in domain_slots
        )
        initializers += [
            self.gen_class_variable(cls, slot, False) for slot in slot_variables
        ]

        return "\n\t".join(initializers)

    def gen_class_variable(
        self, cls: ClassDefinition, slot: SlotDefinition, can_be_positional: bool
    ) -> str:
        """
        Generate a class variable declaration for the supplied slot.  Note: the can_be_positional attribute works,
        but it makes tag/value lists unduly complex, as you can't load them with tag=..., value=... -- you HAVE
        to load positionally. We currently ignore this parameter, meaning that we have a tag/value option for
        any LinkML element

        :param cls: Owning class
        :param slot: slot definition
        :param can_be_positional: True means that positional parameters are allowed.
        :return: Initializer string
        """
        can_be_positional = False  # Force everything to be tag values
        slotname = self.slot_name(slot.name)
        slot_range, default_val = self.range_cardinality(slot, cls, can_be_positional)
        ifabsent_text = (
            ifabsent_value_declaration(slot.ifabsent, self, cls, slot)
            if slot.ifabsent is not None
            else None
        )
        if ifabsent_text is not None:
            default = f"= {ifabsent_text}"
        else:
            default = f"= {default_val}" if default_val else ""
        return f"""{slotname}: {slot_range} {default}"""

    def range_cardinality(
        self,
        slot: SlotDefinition,
        cls: Optional[ClassDefinition],
        positional_allowed: bool,
    ) -> Tuple[str, Optional[str]]:
        """
        Return the range type including initializers, etc.
        Generate a class variable declaration for the supplied slot.  Note: the positional_allowed attribute works,
        but it makes tag/value lists unduly complex, as you can't load them with tag=..., value=... -- you HAVE
        to load positionally. We currently ignore this parameter, meaning that we have a tag/value option for
        any LinkML element

        :param slot: slot to generate type for
        :param cls: containing class -- used to render key slots correctly.  If absent, slot is an add-in
        :param positional_allowed: True Means that we are in the positional space and defaults are not supplied
        :return: python property name and initializer (if any)
        """
        positional_allowed = False  # Force everything to be tag values

        range_type, parent_type, _ = self.class_reference_type(slot, cls)
        pkey = self.class_identifier(slot.range)
        # Special case, inlined, identified range
        if pkey and slot.inlined and slot.multivalued:
            base_key = self.gen_class_reference(
                self.class_identifier_path(slot.range, False)
            )
            num_elements = len(self.schema.classes[slot.range].slots)
            dflt = None if slot.required and positional_allowed else "empty_dict()"
            if num_elements == 1:
                if slot.required:
                    return (
                        f"Union[List[{base_key}], Dict[{base_key}, {range_type}]]",
                        dflt,
                    )
                else:
                    return (
                        f"Optional[Union[List[{base_key}], Dict[{base_key}, {range_type}]]]",
                        dflt,
                    )
            else:
                if slot.required:
                    return (
                        f"Union[Dict[{base_key}, {range_type}], List[{range_type}]]",
                        dflt,
                    )
                else:
                    return (
                        f"Optional[Union[Dict[{base_key}, {range_type}], List[{range_type}]]]",
                        dflt,
                    )

        # All other cases
        if slot.multivalued:
            if slot.required:
                return f"Union[{range_type}, List[{range_type}]]", (
                    None if positional_allowed else "None"
                )
            else:
                return (
                    f"Optional[Union[{range_type}, List[{range_type}]]]",
                    "empty_list()",
                )
        elif slot.required:
            return range_type, (None if positional_allowed else "None")
        else:
            return f"Optional[{range_type}]", "None"

    def class_reference_type(
        self, slot: SlotDefinition, cls: Optional[ClassDefinition]
    ) -> Tuple[str, str, str]:
        """
        Return the type of a slot referencing a class

        :param slot: slot to be typed
        :param cls: owning class.  Used for generating key references
        :return: Python class reference type, most proximal type, most proximal type name
        """
        rangelist = (
            self.class_identifier_path(cls, False)
            if slot.key or slot.identifier
            else self.slot_range_path(slot)
        )
        prox_type = self.slot_range_path(slot)[-1].rsplit(".")[-1]
        prox_type_name = rangelist[-1]

        # Quote forward references - note that enums always gen at the end
        logging.info(f"CHECK: {slot.name} => {slot.range} ")
        if slot.range in self.schema.enums or (
            cls
            and slot.inlined
            and slot.range in self.schema.classes
            and self.forward_reference(slot.range, cls.name)
        ):
            rangelist[-1] = f'"{rangelist[-1]}"'
        return str(self.gen_class_reference(rangelist)), prox_type, prox_type_name

    @staticmethod
    def gen_class_reference(rangelist: List[str]) -> str:
        """
        Return a basic or a union type depending on the number of elements in range list

        :param rangelist: List of types from distal to proximal
        :return:
        """
        base = rangelist[0].rsplit(".")[-1]
        return f"Union[{base}, {rangelist[-1]}]" if len(rangelist) > 1 else base

    def gen_postinits(self, cls: ClassDefinition) -> str:
        """Generate all the typing and existence checks post initialize"""
        post_inits_pre_super = []
        for slot in self.domain_slots(cls):
            if slot.ifabsent:
                dflt = ifabsent_postinit_declaration(slot.ifabsent, self, cls, slot)
                if dflt and dflt != "None":
                    post_inits_pre_super.append(
                        f"if self.{self.slot_name(slot.name)} is None:"
                    )
                    post_inits_pre_super.append(
                        f"\tself.{self.slot_name(slot.name)} = {dflt}"
                    )

        post_inits = []
        if not (cls.mixin or cls.abstract):
            pkeys = self.primary_keys_for(cls)
            for pkey in pkeys:
                slot = self.schema.slots[pkey]
                # TODO: Remove the bypass whenever we get default_range fixed
                if not slot.ifabsent or True:
                    post_inits.append(self.gen_postinit(cls, slot))
        else:
            pkeys = []
        for slot in self.domain_slots(cls):
            if slot.required:
                # TODO: Remove the bypass whenever we get default_range fixed
                if slot.name not in pkeys and (not slot.ifabsent or True):
                    post_inits.append(self.gen_postinit(cls, slot))
        for slot in self.domain_slots(cls):
            if not slot.required:
                # TODO: Remove the bypass whenever we get default_range fixed
                if slot.name not in pkeys and (not slot.ifabsent or True):
                    post_inits.append(self.gen_postinit(cls, slot))

        post_inits_pre_super_line = "\n\t\t".join(
            [p for p in post_inits_pre_super if p]
        ) + ("\n\t\t" if post_inits_pre_super else "")
        post_inits_line = "\n\t\t".join([p for p in post_inits if p])
        return (
            (
                f"""
    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        {post_inits_pre_super_line}{post_inits_line}
        super().__post_init__(**kwargs)"""
                + "\n"
            )
            if post_inits_line or post_inits_pre_super_line
            else ""
        )

    # sort classes such that if C is a child of P then C appears after P in the list
    def _sort_classes(self, clist: List[ClassDefinition]) -> List[ClassDefinition]:
        clist = list(clist)
        slist = []  # sorted
        while len(clist) > 0:
            for i in range(len(clist)):
                candidate = clist[i]
                can_add = False
                if candidate.is_a is None:
                    can_add = True
                else:
                    if candidate.is_a in [p.name for p in slist]:
                        can_add = True
                if can_add:
                    slist = slist + [candidate]
                    del clist[i]
                    break
            if not can_add:
                raise (
                    f"could not find suitable element in {clist} that does not ref {slist}"
                )
        return slist

    def is_key_value_class(self, range_name: DefinitionName) -> bool:
        """
        Return True if range_name references a class with exactly one key and one value

        :param range_name: class definition (name)
        :return: True if meets the special case
        """
        rng = self.schema.classes.get(range_name)
        if rng:
            pkeys = self.primary_keys_for(rng)
            if pkeys:
                return len(rng.slots) - len(pkeys) == 1
        return False

    def gen_postinit(self, cls: ClassDefinition, slot: SlotDefinition) -> Optional[str]:
        """Generate python post init rules for slot in class"""
        rlines: List[str] = []

        if slot.range in self.schema.classes:
            if self.is_class_unconstrained(self.schema.classes[slot.range]):
                return ""

        aliased_slot_name = self.slot_name(
            slot.name
        )  # Mangled name by which the slot is known in python
        range_type, base_type, base_type_name = self.class_reference_type(slot, cls)
        slot_identifier = self.class_identifier(slot.range)

        # Generate existence check for required slots.  Note that inherited classes have to do post init checks because
        # You can't have required elements after optional elements in the parent class
        if slot.required:
            rlines.append(f"if self._is_empty(self.{aliased_slot_name}):")
            rlines.append(f'\tself.MissingRequiredField("{aliased_slot_name}")')

        # Generate the type co-orcion for the various types.
        indent = len(f"self.{aliased_slot_name} = [") * " "
        # NOTE: if you set this to true, we will cast all types.   This may be what we really want
        if not slot.multivalued:
            if slot.required:
                rlines.append(
                    f"if not isinstance(self.{aliased_slot_name}, {base_type_name}):"
                )
            else:
                rlines.append(
                    f"if self.{aliased_slot_name} is not None and "
                    f"not isinstance(self.{aliased_slot_name}, {base_type_name}):"
                )
            # A really wierd case -- a class that has no properties
            if (
                slot.range in self.schema.classes
                and not self.schema.classes[slot.range].slots
            ):
                rlines.append(f"\tself.{aliased_slot_name} = {base_type_name}()")
            else:
                if (
                    (self.class_identifier(slot.range) and not slot.inlined)
                    or slot.range in self.schema.types
                    or slot.range in self.schema.enums
                ):
                    rlines.append(
                        f"\tself.{aliased_slot_name} = {base_type_name}(self.{aliased_slot_name})"
                    )
                else:
                    rlines.append(
                        f"\tself.{aliased_slot_name} = {base_type_name}(**as_dict(self.{aliased_slot_name}))"
                    )
        elif slot.inlined:
            slot_range_cls = self.schema.classes[slot.range]
            identifier = self.class_identifier(slot_range_cls)
            # If we don't have an identifier and we are expecting to be inlined first class elements
            # (inlined_as_list is not True), we will use the first required field as the key.
            #  Note that this may not always work, but the workaround is straight forward -- set inlined_as_list to
            #  True
            if not identifier and not slot.inlined_as_list:
                for range_slot_name in slot_range_cls.slots:
                    range_slot = self.schema.slots[range_slot_name]
                    if range_slot.required:
                        identifier = range_slot.name
                        break
                keyed = False
            else:
                # Place for future expansion
                keyed = True
            if identifier:
                if not slot.inlined_as_list:
                    rlines.append(
                        f'self._normalize_inlined_as_dict(slot_name="{aliased_slot_name}", '
                        f"slot_type={base_type_name}, "
                        f'key_name="{self.aliased_slot_name(identifier)}", '
                        f"keyed={keyed})"
                    )
                else:
                    rlines.append(
                        f'self._normalize_inlined_as_list(slot_name="{aliased_slot_name}", '
                        f"slot_type={base_type_name}, "
                        f'key_name="{self.aliased_slot_name(identifier)}", '
                        f"keyed={keyed})"
                    )
            else:
                # Multivalued, inlined and no identifier
                # TODO: JsonObj([...]) will not be treated correctly here.
                sn = f"self.{aliased_slot_name}"
                rlines.append(f"if not isinstance({sn}, list):")
                rlines.append(f"\t{sn} = [{sn}] if {sn} is not None else []")
                rlines.append(
                    f"{sn} = [v if isinstance(v, {base_type_name}) else {base_type_name}(**as_dict(v)) for v in {sn}]"
                )
        else:
            # Multivalued and not inlined
            # TODO: JsonObj([...]) will fail here as well
            sn = f"self.{aliased_slot_name}"
            rlines.append(f"if not isinstance({sn}, list):")
            rlines.append(f"\t{sn} = [{sn}] if {sn} is not None else []")
            rlines.append(
                f"{sn} = [v if isinstance(v, {base_type_name}) "
                f"else {base_type_name}(v) for v in {sn}]"
            )
        if rlines:
            rlines.append("")
        return "\n\t\t".join(rlines)

    def _slot_iter(
        self,
        cls: ClassDefinition,
        test: Callable[[SlotDefinition], bool],
        first_hit_only: bool = False,
    ) -> Iterator[SlotDefinition]:
        """Return the representation for the set of own slots in cls that pass test

        :param cls: Class containing a set of slots
        :param test: Slot test function
        :param first_hit_only: True means stop on first match.  False means generate all
        :return: Set of slots that match
        """
        for slot in self.all_slots(cls):
            if test(slot):
                yield slot
                if first_hit_only:
                    break

    def primary_keys_for(self, cls: ClassDefinition) -> List[SlotDefinitionName]:
        """Return the primary key for cls.

        Note: At the moment we return at most one entry.  At some point, keys will be expanded to support
              composite keys.

        @param cls: class to get keys for
        @return: List of primary keys or identifiers
        """
        return [
            slot_name
            for slot_name in cls.slots
            if self.schema.slots[slot_name].key
            or self.schema.slots[slot_name].identifier
        ]

    def key_name_for(self, class_name: ClassDefinitionName) -> Optional[str]:
        for slot_name in self.primary_keys_for(self.schema.classes[class_name]):
            return self.formatted_element_name(class_name, True) + camelcase(slot_name)
        return None

    def range_type_name(self, slot: SlotDefinition) -> str:
        """Generate the type name for the slot"""
        cidpath = self.slot_range_path(slot)
        if len(cidpath) < 2:
            return cidpath[0]
        else:
            return f"Union[{cidpath[0]}, {cidpath[-1]}]"

    def forward_reference(self, slot_range: str, owning_class: str) -> bool:
        """Determine whether slot_range is a forward reference"""
        #logging.info(f"CHECKING: {slot_range} {owning_class}")
        if (
            slot_range in self.schema.classes
            and self.schema.classes[slot_range].imported_from
        ) or (
            slot_range in self.schema.enums
            and self.schema.enums[slot_range].imported_from
        ):
            logging.info(f"FALSE: FORWARD: {slot_range} {owning_class} // IMP={self.schema.classes[slot_range].imported_from}")
            return False
        if slot_range in self.schema.enums:
            return True
        for cname in self.schema.classes:
            if cname == owning_class:
                logging.info(f"TRUE: OCCURS SAME: {cname} == {slot_range} owning: {owning_class}")
                return True  # Occurs on or after
            elif cname == slot_range:
                logging.info(f"FALSE: OCCURS BEFORE: {cname} == {slot_range} owning: {owning_class}")
                return False  # Occurs before
        return True

    def python_uri_for(
        self, uriorcurie: Union[str, URIRef]
    ) -> Tuple[str, Optional[str]]:
        """Return the python form of uriorcurie
        :param uriorcurie:
        :return: URI and CURIE form
        """
        ns, ln = self.namespaces.prefix_suffix(uriorcurie)
        if ns == "":
            ns = "DEFAULT_"
        if ns is None:
            return f'"str(uriorcurie)"', None
        return (
            ns.upper() + (f".{ln}" if ln.isidentifier() else f"['{ln}']"),
            ns.upper() + f".curie('{ln}')",
        )

    def gen_slotdefs(self) -> str:
        if self.gen_slots:
            return "class slots:\n\tpass\n\n" + "\n\n".join(
                [
                    self.gen_slot(slot)
                    for slot in self.schema.slots.values()
                    if not slot.imported_from
                ]
            )
        else:
            return ""

    def gen_slot(self, slot: SlotDefinition) -> str:
        python_slot_name = underscore(slot.name)
        slot_uri, slot_curie = self.python_uri_for(slot.slot_uri)
        slot_model_uri, slot_model_curie = self.python_uri_for(
            self.namespaces.uri_or_curie_for(
                self.schema.default_prefix, python_slot_name
            )
        )
        domain = (
            camelcase(slot.domain)
            if slot.domain and not self.schema.classes[slot.domain].mixin
            else "None"
        )
        # Going to omit the range on keys where the domain isn't specified (for now)
        if slot.domain is None and (slot.key or slot.identifier):
            rnge = "URIRef"
        else:
            rnge, _ = self.range_cardinality(
                slot, self.schema.classes[slot.domain] if slot.domain else None, True
            )
        if slot.mappings:
            map_texts = [
                self.namespaces.curie_for(
                    self.namespaces.uri_for(m), default_ok=True, pythonform=True
                )
                for m in slot.mappings
                if m != slot.slot_uri
            ]
        else:
            map_texts = []
        if map_texts:
            mappings = ", mappings = [" + ", ".join(map_texts) + "]"
        else:
            mappings = ""
        pattern = (
            f",\n                   pattern=re.compile(r'{slot.pattern}')"
            if slot.pattern
            else ""
        )
        return f"""slots.{python_slot_name} = Slot(uri={slot_uri}, name="{slot.name}", curie={slot_curie},
                   model_uri={slot_model_uri}, domain={domain}, range={rnge}{mappings}{pattern})"""

    def gen_enumerations(self) -> str:
        return "\n\n".join(
            [
                self.gen_enum(enum)
                for enum in self.schema.enums.values()
                if not enum.imported_from
            ]
        )

    def gen_enum(self, enum: EnumDefinition) -> str:
        enum_name = camelcase(enum.name)
        return f"""
class {enum_name}(EnumDefinitionImpl):
    {self.gen_enum_comment(enum)}
    {self.gen_enum_description(enum, enum_name)}
""".strip()

    def gen_enum_comment(self, enum: EnumDefinition) -> str:
        return (
            f'"""\n\t{wrapped_annotation(be(enum.description))}\n\t"""'
            if be(enum.description)
            else ""
        )

    def gen_enum_description(self, enum: EnumDefinition, enum_name: str) -> str:
        return f"""
    {self.gen_pvs(enum)}

    {self.gen_enum_definition(enum, enum_name)}
    {self.gen_pvs2(enum)}
""".strip()

    def gen_pvs(self, enum: EnumDefinition) -> str:
        """
        Generate the python compliant permissible value initializers as a set of class variables
        @param enum:
        @return:
        """
        init_list = []
        for pv in enum.permissible_values.values():
            if str.isidentifier(pv.text) and not keyword.iskeyword(pv.text):
                l1 = f"{pv.text} = "
                l1len = len(l1)
                l2ton = "\n" + l1len * " "
                init_list.append(l1 + (l2ton.join(self.gen_pv_constructor(pv, l1len))))
        return "\n\t".join(init_list).strip()

    def gen_enum_definition(self, enum: EnumDefinition, enum_name: str) -> str:
        enum_desc = (
            enum.description.replace('"', '\\"').replace(r"\n", r"\\n")
            if enum.description
            else None
        )
        desc = f'\t\tdescription="{enum_desc}",\n' if enum.description else ""
        cs = (
            f"\t\tcode_set={self.namespaces.curie_for(self.namespaces.uri_for(enum.code_set), default_ok=False, pythonform=True)},\n"
            if enum.code_set
            else ""
        )
        tag = f'\t\tcode_set_tag="{enum.code_set_tag}",\n' if enum.code_set_tag else ""
        ver = (
            f'\t\tcode_set_version="{enum.code_set_version}",\n'
            if enum.code_set_version
            else ""
        )
        vf = (
            f"\t\tpv_formula=PvFormulaOptions.{enum.pv_formula.code.text},\n"
            if enum.pv_formula
            else ""
        )

        return f"""_defn = EnumDefinition(\n\t\tname="{enum_name}",\n{desc}{cs}{tag}{ver}{vf}\t)"""

    def gen_pvs2(self, enum: EnumDefinition) -> str:
        """
        Generate the non-python compliant permissible value initializers as a set of setattr instructions
        @param enum:
        @return:
        """
        if any(
            not str.isidentifier(pv.text) or keyword.iskeyword(pv.text)
            for pv in enum.permissible_values.values()
        ):
            return f"""
    @classmethod
    def _addvals(cls):
        {self.gen_pvs2_initializers(enum)}"""
        else:
            return ""

    def gen_pvs2_initializers(self, enum: EnumDefinition) -> str:
        init_list = []
        for pv in enum.permissible_values.values():
            if not str.isidentifier(pv.text) or keyword.iskeyword(pv.text):
                l1 = "        setattr("
                l2ton = len(l1) * " "
                pv_cons = "\n".join(self.gen_pv_constructor(pv, len(l1)))
                pv_text = pv.text.replace('"', '\\"').replace(r"\n", r"\\n")
                init_list.append(f'{l1}cls, "{pv_text}",\n{l2ton}{pv_cons} )')
        return "\n".join(init_list).strip()

    def gen_pv_constructor(self, pv: PermissibleValue, indent: int) -> List[str]:
        """
        Generate a permissible value constructor
        @param pv: Value to be constructed
        @param indent: number of additional spaces to add on successive lines
        @return: Permissible value constructor
        """
        # PermissibleValue(text="CODE",
        #                  description="...",
        #                  meaning="...")
        constructor = "PermissibleValue("
        indent = (len(constructor) + indent) * " "
        c1 = "," if pv.description or pv.meaning else ")"
        rval = [f'{constructor}text="{pv.text}"{c1}']
        if pv.description:
            c2 = "," if pv.meaning else ")"
            rval.append(f'{indent}description="{pv.description}"{c2}')
        if pv.meaning:
            pv_meaning = self.namespaces.curie_for(
                self.namespaces.uri_for(pv.meaning), default_ok=False, pythonform=True
            )
            rval.append(f"{indent}meaning={pv_meaning})")
        return rval


@shared_arguments(PythonGenerator)
@click.command()
@click.option(
    "--head/--no-head", default=True, show_default=True, help="Emit metadata heading"
)
@click.option(
    "--genmeta/--no-genmeta",
    default=False,
    show_default=True,
    help="Generating metamodel. Only use this for generating meta.py",
)
@click.option(
    "--classvars/--no-classvars",
    default=True,
    show_default=True,
    help="Generate CLASSVAR info",
)
@click.option(
    "--slots/--no-slots",
    default=True,
    show_default=True,
    help="Generate Slot information",
)
@click.option(
    "--validate/--no-validate",
    default=False,
    show_default=True,
    help="Validate generated code by compiling it",
)
@click.version_option(__version__, "-V", "--version")
def cli(yamlfile, head=True, genmeta=False, classvars=True, slots=True, validate=False, **args):
    """Generate python classes to represent a LinkML model"""
    gen = PythonGenerator(
            yamlfile,
            emit_metadata=head,
            genmeta=genmeta,
            gen_classvars=classvars,
            gen_slots=slots,
            **args,
        )
    if validate:
        mod = gen.compile_module()
        logging.info(f"Module {mod} compiled successfully")
    print(
        gen.serialize(emit_metadata=head, **args)
    )


if __name__ == "__main__":
    cli()



import logging
import os
from collections import defaultdict
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, List, Optional, TextIO, Union, cast

import click
from jinja2 import Template
from linkml_runtime.linkml_model.meta import (ClassDefinition,
                                              ClassDefinitionName, Element,
                                              EnumDefinitionName, Prefix,
                                              SchemaDefinition, SlotDefinition,
                                              SlotDefinitionName,
                                              TypeDefinition,
                                              TypeDefinitionName)
from linkml_runtime.utils.formatutils import camelcase, underscore
from linkml_runtime.utils.schemaview import SchemaView

from linkml._version import __version__
from linkml.utils.generator import Generator, shared_arguments

template = """
{% for pfxn, pfx in schema.prefixes.items() -%}
PREFIX {{pfxn}}: <{{pfx.prefix_reference}}>
{% endfor %}

{% for cn, c in schema.classes.items() if not c.mixin and not c.abstract %}

## --
## Checks for {{ cn }}
## --

# @CHECK permitted_{{cn}}
SELECT ?g ?s ?p WHERE {
 GRAPH ?g {
  ?s rdf:type {{ schema_view.get_uri(cn) }} ;
     ?p ?o .
  FILTER ( ?p NOT IN (
   {% for sn in schema_view.class_slots(cn) -%}
   {{ schema_view.get_uri(schema_view.get_slot(sn, attributes=True)) }},
   {% endfor -%}
   rdf:type ))
 }
 {{ extra }}
} {{ limit }}


{% for slot in schema_view.class_induced_slots(cn) -%}

{% if slot.required %}
# @CHECK required_{{cn}}_{{slot.name}}
SELECT
  ?check
  ?graph
  ?subject
  ?predicate WHERE {
 GRAPH ?graph {
  ?subject rdf:type {{ schema_view.get_uri(cn) }} .
  FILTER NOT EXISTS { ?subject  {{ schema_view.get_uri(slot) }} ?o  }
 }
 VALUES ?check { linkml:required }
 VALUES ?predicate { {{schema_view.get_uri(slot)}} }
 {{ extra }}
}  {{ limit }}
{% endif %}

{% if slot.range in schema_view.all_class() %}
# @CHECK object_range_{{cn}}_{{slot.name}}
SELECT
  ?check
  ?graph
  ?subject
  ?predicate
  ?object
WHERE {
 GRAPH ?graph {
  ?subject rdf:type {{ schema_view.get_uri(cn) }} ;
     ?predicate ?object .
  FILTER NOT EXISTS {
    ?object rdf:type ?otype .
    FILTER ( ?otype IN (
      {% for a in schema_view.class_descendants(slot.range) -%}
      {{ schema_view.get_uri(a) }}
      {{ ", " if not loop.last else "" }}
      {% endfor -%} ))
  }
 }
 VALUES ?check { linkml:range }
 VALUES ?predicate { {{ schema_view.get_uri(slot) }}  }
 {{ extra }}
}  {{ limit }}
{% endif %}

{%- endfor %}


## -- End of checks for {{ cn }}
{% endfor %}
"""

x = """
{% for sn in schema_view.class_slots(c.name) %}
     {{ schema.slots[sn].slot_uri }}
   {% endfor %}
"""


def materialize_schema(schemaview: SchemaView):
    schema = schemaview.schema
    if "rdf" not in schema.prefixes:
        schema.prefixes["rdf"] = Prefix(
            "rdf", "http://www.w3.org/1999/02/22-rdf-syntax-ns#"
        )
    for scn in schemaview.imports_closure():
        for pfxn, pfx in schemaview.schema_map[scn].prefixes.items():
            if pfxn not in schema:
                schema.prefixes[pfxn] = pfx
    for cn, c in schemaview.all_class().items():
        for a in list(c.attributes.values()):
            schema.slots[a.name] = a
            c.slots.append(a.name)
            del c.attributes[a.name]
    schemaview.set_modified()
    for cn, c in schemaview.all_class().items():
        for s in schemaview.class_induced_slots(cn):
            if s.name not in c.slots:
                c.slots.append(s.name)
            c.slot_usage[s.name] = s
            s.slot_uri = schemaview.get_uri(s)


@dataclass
class SparqlGenerator(Generator):
    """
    Generates SPARQL queries that can be used for delayed validation
    """

    # ClassVars
    generatorname = os.path.basename(__file__)
    valid_formats = ["sparql"]
    visit_all_class_slots = False
    uses_schemaloader = False

    # ObjectVars
    named_graphs: Optional[List[str]] = None
    limit: Optional[int] = None
    sparql: Optional[str] = None

    def __post_init__(self):
        self.schemaview = SchemaView(self.schema)
        materialize_schema(self.schemaview)
        super().__post_init__()
        self.queries = self.generate_sparql(named_graphs=self.named_graphs, limit=self.limit)

    def generate_sparql(self, named_graphs=None, limit: int = None):
        template_obj = Template(template)
        extra = ""
        if named_graphs is not None:
            extra += f'FILTER( ?graph in ( {",".join(named_graphs)} ))'
        logging.info(f"Named Graphs = {named_graphs} // extra={extra}")
        if limit is not None and isinstance(limit, int):
            limit = f"LIMIT {limit}"
        else:
            limit = ""
        sparql = template_obj.render(
            schema_view=self.schemaview, schema=self.schema, limit=limit, extra=extra
        )
        self.sparql = sparql
        queries = self.split_sparql(sparql)
        return queries

    def serialize(self, directory=None) -> str:
        if directory is not None:
            Path(directory).mkdir(parents=True, exist_ok=True)
            for qn, q in self.queries.items():
                qpath = os.path.join(directory, f"{qn}.rq")
                with open(qpath, "w", encoding="UTF-8") as stream:
                    stream.write(q)
        return self.sparql

    def split_sparql(self, sparql: str) -> Dict[str, str]:
        lines = sparql.split("\n")
        prolog = ""
        queries = defaultdict(str)
        q = None
        for line in lines:
            if line.startswith("# @"):
                q = underscore(line.replace("# @", ""))
                queries[q] = prolog + "\n"
            elif q is None:
                if line.lower().startswith("prefix"):
                    prolog += line + "\n"
            else:
                queries[q] += line + "\n"
        return queries


@shared_arguments(SparqlGenerator)
@click.command()
@click.option("--dir", "-d", help="Directory in which queries will be deposited")
@click.version_option(__version__, "-V", "--version")
def cli(yamlfile, dir, **kwargs):
    """Generate SPARQL queries for validation

    This will generate a directory of queries that cann be used for QC over a triplestore that
    is conformant to the same LinkML schema.

    Each query in the directory will be of the form

        CHECK_<ConstraintType>_<SchemaElement>.rq

    Example:

        gen-sparql -d ./sparql/ personinfo.yaml
    """
    SparqlGenerator(yamlfile, **kwargs).serialize(directory=dir)


if __name__ == "__main__":
    cli()



import json
import logging
import os
from collections import UserDict
from copy import deepcopy
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional, Tuple, Union

import click
from linkml_runtime.linkml_model.meta import (AnonymousClassExpression,
                                              AnonymousSlotExpression,
                                              ClassDefinition, EnumDefinition,
                                              PermissibleValue,
                                              PermissibleValueText,
                                              SlotDefinition, ClassDefinitionName,
                                              metamodel_version)
from linkml_runtime.utils.formatutils import be, camelcase, underscore

from linkml._version import __version__
from linkml.utils.generator import Generator, shared_arguments

# Map from underlying python data type to json equivalent
# Note: The underlying types are a union of any built-in python datatype + any type defined in
#       linkml-runtime/utils/metamodelcore.py
# Note the keys are all lower case
json_schema_types: Dict[str, Tuple[str, Optional[str]]] = {
    "int": ("integer", None),
    "integer": ("integer", None),
    "bool": ("boolean", None),
    "boolean": ("boolean", None),
    "float": ("number", None),
    "double": ("number", None),
    "decimal": ("number", None),
    "xsddate": ("string", "date"),
    "xsddatetime": ("string", "date-time"),
    "xsdtime": ("string", "time"),
}

class JsonSchema(UserDict):

    OPTIONAL_IDENTIFIER_SUFFIX = "__identifier_optional"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._lax_forward_refs = {}

    def add_def(self, name: str, subschema: "JsonSchema") -> None:
        canonical_name = camelcase(name)

        if '$defs' not in self:
            self['$defs'] = {}

        if 'title' not in subschema:
            subschema['title'] = canonical_name

        self['$defs'][canonical_name] = subschema

        if canonical_name in self._lax_forward_refs:
            identifier_name = self._lax_forward_refs.pop(canonical_name)
            self.add_lax_def(canonical_name, identifier_name)

    def add_lax_def(self, name: str, identifier_name: str) -> None:
        # JSON-Schema does not have inheritance,
        # so we duplicate slots from inherited parents and mixins
        # Maps e.g. Person --> Person__identifier_optional
        # for use when Person is a range of an inlined-as-dict slot
        canonical_name = camelcase(name)

        if "$defs" not in self or canonical_name not in self["$defs"]:
            self._lax_forward_refs[canonical_name] = identifier_name
        else:
            lax_cls = deepcopy(self["$defs"][canonical_name])
            lax_cls['required'].remove(identifier_name)
            self["$defs"][canonical_name + self.OPTIONAL_IDENTIFIER_SUFFIX] = lax_cls

    def add_property(self, name: str, subschema: "JsonSchema", required: bool = False) -> None:
        canonical_name = underscore(name)
        
        if 'properties' not in self:
            self['properties'] = {}
        
        self['properties'][canonical_name] = subschema

        if required:
            if 'required' not in self:
                self['required'] = []
            
            self['required'].append(canonical_name)

    def add_keyword(self, keyword: str, value: Any, applies_to_all_array_elements: bool = False):
        if value is None:
            return
        
        if applies_to_all_array_elements and self.is_array:
            self['items'][keyword] = value
        else:
            self[keyword] = value

    @property
    def is_array(self):
        return self.get('type') == 'array'

    @property
    def is_object(self):
        return self.get('type') == 'object'

    def to_json(self, **kwargs) -> str:
        return json.dumps(self.data, default=lambda d: d.data, **kwargs)

    @classmethod
    def ref_for(cls, class_name: Union[str, List[str]], identifier_optional: bool = False):
        def _ref(class_name):
            return JsonSchema({ 
                "$ref": f"#/$defs/{camelcase(class_name)}{cls.OPTIONAL_IDENTIFIER_SUFFIX if identifier_optional else ''}"
            })
        
        if isinstance(class_name, list):
            return JsonSchema({
                "anyOf": [_ref(name) for name in class_name]
            })
        else:
            return _ref(class_name)

    @classmethod
    def array_of(cls, subschema: "JsonSchema") -> "JsonSchema":
        schema = {
            "type": "array",
            "items": subschema
        }

        return JsonSchema(schema)


@dataclass
class JsonSchemaGenerator(Generator):
    """
    Generates JSONSchema documents from a LinkML SchemaDefinition

    - Each linkml class generates a schema
    - inheritance hierarchies are rolled-down from ancestors
    - Composition not yet implemented
    - Enumerations treated as strings
    - Foreign key references are treated as semantics-free strings
    """

    # ClassVars
    generatorname = os.path.basename(__file__)
    generatorversion = "0.0.3"
    valid_formats = ["json"]
    uses_schemaloader = False

    #@deprecated("Use top_class")
    topClass: Optional[str] = None

    not_closed: Optional[bool] = field(default_factory=lambda: True)
    """If not closed, then an open-ended set of attributes can be instantiated for any object"""

    indent: int = 4

    inline: bool = False
    top_class: Optional[ClassDefinitionName] = None  ## JSON object is one instance of this
    """Class instantiated by the root node of the document tree"""

    include_range_class_descendants: bool = field(default_factory=lambda: False)

    top_level_schema: JsonSchema = None

    def __post_init__(self):
        if self.topClass:
            logging.warning(f"topClass is deprecated - use top_class")
            self.top_class = self.topClass

        super().__post_init__()

    def start_schema(self, inline: bool = False) -> JsonSchema:
        self.inline = inline

        self.top_level_schema = JsonSchema({
            "$schema": "http://json-schema.org/draft-07/schema#",
            "$id": self.schema.id,
            "metamodel_version": metamodel_version,
            "version": self.schema.version if self.schema.version else None,
            "title": self.schema.name,
            "type": "object",
            "additionalProperties": self.not_closed,
        })

    def handle_class(self, cls: ClassDefinition) -> None:
        if cls.mixin or cls.abstract:
            return
        
        additional_properties = False
        if self.is_class_unconstrained(cls):
            additional_properties = True
        
        class_subschema = JsonSchema({
            "type": "object",
            "additionalProperties": additional_properties,
            "description": be(cls.description),
        })

        for slot_definition in self.schemaview.class_induced_slots(cls.name):
            self.handle_class_slot(
                subschema=class_subschema,
                cls=cls,
                slot=slot_definition
            )

        rule_subschemas = []
        for rule in cls.rules:
            subschema = JsonSchema()

            open_world = rule.open_world
            if open_world is None:
                open_world = False

            if_subschema = self.get_subschema_for_anonymous_class(rule.preconditions, properties_required=True)
            if if_subschema:
                subschema["if"] = if_subschema
            
            then_subschema = self.get_subschema_for_anonymous_class(rule.postconditions, properties_required=not open_world)
            if then_subschema:
                subschema["then"] = then_subschema

            # same as required requirements as postconditions?
            else_subschema = self.get_subschema_for_anonymous_class(rule.elseconditions, properties_required=not open_world)
            if else_subschema:
                subschema["else"] = else_subschema

            rule_subschemas.append(subschema)

            if rule.bidirectional:
                inverse_subschema = JsonSchema()

                if then_subschema:
                    inverse_subschema["if"] = then_subschema
                
                if if_subschema:
                    inverse_subschema["then"] = if_subschema

                rule_subschemas.append(inverse_subschema)

        if len(rule_subschemas) == 1:
            self.top_level_schema.update(rule_subschemas[0])
        elif len(rule_subschemas) > 1:
            self.top_level_schema["allOf"] = rule_subschemas

        self.top_level_schema.add_def(cls.name, class_subschema)


    def get_subschema_for_anonymous_class(self, cls: AnonymousClassExpression, properties_required: bool = False) -> Union[None, JsonSchema]:
        if not cls:
            return None

        subschema = JsonSchema()
        for slot in cls.slot_conditions.values():
            prop = self.get_subschema_for_slot(slot, omit_type=True)
            subschema.add_property(self.aliased_slot_name(slot), prop, properties_required)

        if cls.any_of is not None and len(cls.any_of) > 0:
            subschema["anyOf"] = [self.get_subschema_for_anonymous_class(c, properties_required) for c in cls.any_of]

        if cls.all_of is not None and len(cls.all_of) > 0:
            subschema["allOf"] = [self.get_subschema_for_anonymous_class(c, properties_required) for c in cls.all_of]

        if cls.exactly_one_of is not None and len(cls.exactly_one_of) > 0:
            subschema["oneOf"] = [self.get_subschema_for_anonymous_class(c, properties_required) for c in cls.exactly_one_of]

        if cls.none_of is not None and len(cls.none_of) > 0:
            subschema["not"] = {
                "anyOf": [self.get_subschema_for_anonymous_class(c, properties_required) for c in cls.any_of]
            }

        return subschema


    def handle_enum(self, enum: EnumDefinition) -> None:
        # TODO: this only works with explicitly permitted values. It will need to be extended to
        # support other pv_formula

        def extract_permissible_text(pv):
            if type(pv) is str:
                return pv
            if type(pv) is PermissibleValue:
                return pv.text.code
            if type(pv) is PermissibleValueText:
                return pv
            raise ValueError(f"Invalid permissible value in enum {enum}: {pv}")

        permissible_values_texts = list(
            map(extract_permissible_text, enum.permissible_values or [])
        )

        enum_schema = JsonSchema({
            "type": "string",
            "enum": permissible_values_texts,
            "description": be(enum.description),
        })
        self.top_level_schema.add_def(enum.name, enum_schema)

    def get_type_info_for_slot_subschema(self, slot: AnonymousSlotExpression, parent_slot_is_inlined: bool) -> Tuple[str, str, Union[str, List[str]]]:
        typ = None  # JSON Schema type (https://json-schema.org/understanding-json-schema/reference/type.html)
        reference = None  # Reference to a JSON schema entity (https://json-schema.org/understanding-json-schema/structuring.html#ref)
        fmt = None  # JSON Schema format (https://json-schema.org/understanding-json-schema/reference/string.html#format)

        slot_is_inlined = self.schemaview.is_inlined(slot)
        if slot.range in self.schemaview.all_types().keys():
            schema_type = self.schemaview.induced_type(slot.range)
            (typ, fmt) = json_schema_types.get(schema_type.base.lower(), ("string", None))
        elif slot.range in self.schemaview.all_enums().keys():
            reference = slot.range
        elif (slot_is_inlined or parent_slot_is_inlined) and slot.range in self.schemaview.all_classes().keys():
            descendants = [desc for desc in self.schemaview.class_descendants(slot.range) 
                if not self.schemaview.get_class(desc).abstract]
            if descendants and self.include_range_class_descendants:
                reference = descendants
            else:
                reference = slot.range
        else:
            typ = "string"

        return (typ, fmt, reference)

    def get_subschema_for_slot(self, slot: SlotDefinition, omit_type: bool = False) -> JsonSchema:
        slot_has_range_union = slot.any_of is not None and len(slot.any_of) > 0 and all(s.range is not None for s in slot.any_of)
        if omit_type:
            prop = JsonSchema()
        else:
            slot_is_inlined = self.schemaview.is_inlined(slot)

            if slot_has_range_union:
                items = []
                for sub_slot in slot.any_of:
                    typ, fmt, reference = self.get_type_info_for_slot_subschema(sub_slot, slot_is_inlined)
                    if reference is not None:
                        item = JsonSchema.ref_for(reference)
                    elif fmt is None:
                        item = JsonSchema({"type": typ})
                    else:
                        item = JsonSchema({"type": typ, "format": fmt})
                    items.append(item)
                subschema = JsonSchema({
                    "anyOf": items
                })
                if slot.multivalued:
                    prop = JsonSchema.array_of(subschema)
                else:
                    prop = subschema
            else:
                typ, fmt, reference = self.get_type_info_for_slot_subschema(slot, slot_is_inlined)
                if slot_is_inlined:
                    # If inline we have to include redefined slots
                    if slot.multivalued:
                        range_id_slot = self.schemaview.get_identifier_slot(slot.range, use_key=True)
                        if range_id_slot is not None and not slot.inlined_as_list:
                            prop = JsonSchema({
                                "type": "object",
                                "additionalProperties": JsonSchema.ref_for(reference, identifier_optional=True)
                            })
                            self.top_level_schema.add_lax_def(reference, self.aliased_slot_name(range_id_slot))
                        else:
                            prop = JsonSchema.array_of(JsonSchema.ref_for(reference))
                    else:
                        prop = JsonSchema.ref_for(reference)
                else:
                    if slot.multivalued:
                        if reference is not None:
                            prop = JsonSchema.array_of(JsonSchema.ref_for(reference))
                        elif fmt is None:
                            prop = JsonSchema.array_of(JsonSchema({"type": typ}))
                        else:
                            prop = JsonSchema.array_of(JsonSchema({"type": typ, "format": fmt}))
                    else:
                        if reference is not None:
                            prop = JsonSchema.ref_for(reference)
                        elif fmt is None:
                            prop = JsonSchema({"type": typ})
                        else:
                            prop = JsonSchema({"type": typ, "format": fmt})

        prop.add_keyword('description', slot.description)

        prop.add_keyword('pattern', slot.pattern, applies_to_all_array_elements=True)
        prop.add_keyword('minimum', slot.minimum_value, applies_to_all_array_elements=True)
        prop.add_keyword('maximum', slot.maximum_value, applies_to_all_array_elements=True)
        prop.add_keyword('const', slot.equals_string, applies_to_all_array_elements=True)
        prop.add_keyword('const', slot.equals_number, applies_to_all_array_elements=True)

        if prop.is_array:
            prop.add_keyword('minItems', slot.minimum_cardinality)
            prop.add_keyword('maxItems', slot.maximum_cardinality)

        if prop.is_object:
            prop.add_keyword('minProperties', slot.minimum_cardinality)
            prop.add_keyword('maxProperties', slot.maximum_cardinality)

        if slot.any_of is not None and len(slot.any_of) > 0:
            if not slot_has_range_union:
                prop['anyOf'] = [self.get_subschema_for_slot(s, omit_type=True) for s in slot.any_of]

        if slot.all_of is not None and len(slot.all_of) > 0:
            prop['allOf'] = [self.get_subschema_for_slot(s, omit_type=True) for s in slot.all_of]

        if slot.exactly_one_of is not None and len(slot.exactly_one_of) > 0:
            prop['oneOf'] = [self.get_subschema_for_slot(s, omit_type=True) for s in slot.exactly_one_of]

        if slot.none_of is not None and len(slot.none_of) > 0:
            prop['not'] = {
                'anyOf': [self.get_subschema_for_slot(s, omit_type=True) for s in slot.none_of]
            }

        return prop


    def handle_class_slot(self, subschema: JsonSchema, cls: ClassDefinition, slot: SlotDefinition) -> None:
        class_id_slot = self.schemaview.get_identifier_slot(cls.name, use_key=True)
        slot_is_required = slot.required or slot == class_id_slot

        aliased_slot_name = self.aliased_slot_name(slot)
        prop = self.get_subschema_for_slot(slot)
        subschema.add_property(aliased_slot_name, prop, slot_is_required)

        if (
            self.top_class is not None and camelcase(self.top_class) == camelcase(cls.name)
        ) or (self.top_class is None and cls.tree_root):
            self.top_level_schema.add_property(aliased_slot_name, prop, slot_is_required)

    def serialize(self, **kwargs) -> str:
        self.start_schema()
        for enum_definition in self.schemaview.all_enums().values():
            self.handle_enum(enum_definition)

        for class_definition in self.schemaview.all_classes().values():
            self.handle_class(class_definition)

        return self.top_level_schema.to_json(sort_keys=True, indent=self.indent if self.indent > 0 else None)


@shared_arguments(JsonSchemaGenerator)
@click.command()
@click.option(
    "-i",
    "--inline",
    is_flag=True,
    help="""
Generate references to types rather than inlining them.
Note that declaring a slot as inlined: true will always inline the class
""",
)
@click.option(
    "-t",
    "--top-class",
    help="""
Top level class; slots of this class will become top level properties in the json-schema
""",
)
@click.option(
    "--not-closed/--closed",
    default=True,
    show_default=True,
    help="""
Set additionalProperties=False if closed otherwise true if not closed at the global level
""",
)
@click.option(
    "--include-range-class-descendants/--no-range-class-descendants",
    default=False,
    show_default=False,
    help="""
When handling range constraints, include all descendants of the range class instead of just the range class
""",
)
@click.option(
    "--indent",
    default=4,
    show_default=True,
    help="""
If this is a positive number the resulting JSON will be pretty-printed with that indent level. Set to 0 to
disable pretty-printing and return the most compact JSON representation
"""
)
@click.version_option(__version__, "-V", "--version")
def cli(yamlfile, **kwargs):
    """Generate JSON Schema representation of a LinkML model"""
    print(JsonSchemaGenerator(yamlfile, **kwargs).serialize(**kwargs))


if __name__ == "__main__":
    cli()



import os
from dataclasses import dataclass, field
from typing import TextIO, Union

import click
from linkml_runtime.linkml_model.meta import (ClassDefinition,
                                              SchemaDefinition, SlotDefinition)
from linkml_runtime.utils.formatutils import camelcase, lcamelcase

from linkml._version import __version__
from linkml.utils.generator import Generator, shared_arguments


@dataclass
class ProtoGenerator(Generator):
    """
    A `Generator` for creating Protobuf schemas from a linkml schema.

    """

    # ClassVars
    generatorname = os.path.basename(__file__)
    generatorversion = "0.1.1"
    valid_formats = ["proto"]
    visit_all_class_slots = True
    uses_schemaloader = True

    # ObjectVars
    relative_slot_num: int = field(default_factory=lambda: 0)

    def __post_init__(self):
        super().__post_init__()
        self.generate_header()

    def generate_header(self):
        print(f' syntax="proto3";')
        print(f" package")
        print(f"// metamodel_version: {self.schema.metamodel_version}")
        if self.schema.version:
            print(f"// version: {self.schema.version}")

    def visit_class(self, cls: ClassDefinition) -> bool:
        if cls.mixin or cls.abstract or not cls.slots:
            return False
        if cls.description:
            for dline in cls.description.split("\n"):
                print(f"// {dline}")
        print(f"message {camelcase(cls.name)}")
        print(" {")
        self.relative_slot_num = 0
        return True

    def end_class(self, cls: ClassDefinition) -> None:
        print(" }")

    def visit_class_slot(
        self, cls: ClassDefinition, aliased_slot_name: str, slot: SlotDefinition
    ) -> None:
        qual = "repeated " if slot.multivalued else ""
        slotname = lcamelcase(aliased_slot_name)
        slot_range = camelcase(slot.range)
        if slot.rank is None:
            # numbering of slots is important in the proto implementation
            # and should be determined by the rank param.
            slot.rank = 0
        print(f" {qual} {lcamelcase(slot_range)} {(slotname)} = {slot.rank}")


@shared_arguments(ProtoGenerator)
@click.version_option(__version__, "-V", "--version")
@click.command()
def cli(yamlfile, **args):
    """Generate proto representation of LinkML model"""
    print(ProtoGenerator(yamlfile, **args).serialize(**args))


if __name__ == "__main__":
    cli()



"""
Generators translate between a SchemaDefinition and an alternative
representation such as JsonSchema
"""
# Generator version numbers
__all__ = [
    "csvgen",
    "dotgen",
    "docgen",
    "golrgen",
    "graphqlgen",
    "javagen",
    "jsonldcontextgen",
    "jsonldgen",
    "jsonschemagen",
    "markdowngen",
    "namespacegen",
    "owlgen",
    "protogen",
    "pythongen",
    "rdfgen",
    "shexgen",
    "sssomgen",
    "summarygen",
    "yamlgen",
    "yumlgen",
    "PYTHON_GEN_VERSION",
]
GENERATOR_BASE = "0.9"

PYTHON_GEN_VERSION = GENERATOR_BASE + ".0"
JAVA_GEN_VERSION = GENERATOR_BASE + ".0"
